from psynet.modular_page import (
    DropdownControl,
    ModularPage,
    NumberControl,
    PushButtonControl,
    RadioButtonControl,
    TextControl,
)
from psynet.timeline import FailedValidation, Module, conditional, join
from psynet.utils import (
    get_country_dict,
    get_language_dict,
    get_locale,
    get_logger,
    get_translator,
)

logger = get_logger()


class BasicDemography(Module):
    def __init__(
        self,
        label="basic_demography",
    ):
        self.label = label
        self.elts = join(
            Gender(),
            Age(),
            CountryOfBirth(),
            CountryOfResidence(),
            FormalEducation(),
        )
        super().__init__(self.label, self.elts)


class Language(Module):
    def __init__(
        self,
        label="language",
    ):
        self.label = label
        self.elts = join(
            MotherTongue(),
            MoreThanOneLanguage(),
            conditional(
                "more_than_one_language",
                lambda experiment, participant: participant.answer == "yes",
                LanguagesInOrderOfProficiency(),
            ),
        )
        super().__init__(self.label, self.elts)


class BasicMusic(Module):
    def __init__(
        self,
        label="basic_music",
    ):
        self.label = label
        self.elts = join(
            YearsOfFormalTraining(),
            HoursOfDailyMusicListening(),
            MoneyFromPlayingMusic(),
        )
        super().__init__(self.label, self.elts)


class Dance(Module):
    def __init__(
        self,
        label="dance",
    ):
        self.label = label
        self.elts = join(
            DanceSociallyOrProfessionally(),
            conditional(
                "dance_socially_or_professionally",
                lambda experiment, participant: (
                    participant.answer in ["socially", "professionally"]
                ),
                LastTimeDanced(),
            ),
        )
        super().__init__(self.label, self.elts)


class SpeechDisorders(Module):
    def __init__(
        self,
        label="speech_disorders",
    ):
        self.label = label
        self.elts = join(
            SpeechLanguageTherapy(),
            DiagnosedWithDyslexia(),
        )
        super().__init__(self.label, self.elts)


class Income(Module):
    def __init__(
        self,
        label="income",
    ):
        self.label = label
        self.elts = join(
            HouseholdIncomePerYear(),
        )
        super().__init__(self.label, self.elts)


class ExperimentFeedback(Module):
    def __init__(
        self,
        label="feedback",
    ):
        self.label = label
        self.elts = join(
            LikedExperiment(),
            FoundExperimentDifficult(),
            EncounteredTechnicalProblems(),
        )
        super().__init__(self.label, self.elts)


# Basic demography #
class Gender(ModularPage):
    def __init__(
        self,
        label="gender",
    ):
        _p = get_translator(context=True)
        prompt = _p("gender", "How do you identify yourself?")
        self.label = label
        self.prompt = prompt
        self.time_estimate = 5

        control = RadioButtonControl(
            ["female", "male", "non_binary", "not_specified", "prefer_not_to_say"],
            [
                _p("gender", "Female"),
                _p("gender", "Male"),
                _p("gender", "Non-binary"),
                _p("gender", "Not specified"),
                _p("gender", "I prefer not to answer"),
            ],
            name="gender",
            show_free_text_option=True,
            placeholder_text_free_text=_p("gender", "Specify yourself"),
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


class Age(ModularPage):
    def __init__(
        self,
        label="age",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p("age", "What is your age?")
        self.time_estimate = 5
        super().__init__(
            self.label,
            self.prompt,
            control=NumberControl(),
            time_estimate=self.time_estimate,
            save_answer=label,
        )

    def validate(self, response, **kwargs):
        _p = get_translator(context=True)
        answer = response.answer
        error_msg = (
            _p("age", "You need to provide your age as an integer between 0 and 120!")
            + " "
            + _p("age", "Your answer was: '{AGE}'").format(AGE=answer)
        )
        try:
            age = int(answer)
            if not (0 < age < 120):
                return FailedValidation(error_msg)
            else:
                return None
        except ValueError:
            return FailedValidation(error_msg)


class CountryDropdown(ModularPage):
    def __init__(self, label):
        self.label = label
        _p = get_translator(context=True)
        self.time_estimate = 5
        locale = get_locale()
        country_dict = get_country_dict(locale)
        control = DropdownControl(
            choices=list(country_dict.keys()) + ["OTHER"],
            labels=list(country_dict.values())
            + [_p("country-select", "Other country")],
            default_text=_p("country-select", "Select a country"),
            name=self.label,
        )
        super().__init__(
            self.label,
            self.get_prompt(),
            control=control,
            time_estimate=self.time_estimate,
            save_answer="country",
        )

    def get_prompt(self):
        raise NotImplementedError()

    def validate(self, response, **kwargs):
        _p = get_translator(context=True)
        if self.control.force_selection and response.answer == "":
            return FailedValidation(
                _p("country-select", "You need to select a country!")
            )
        return None


class CountryOfBirth(CountryDropdown):
    def __init__(
        self,
        label="country_of_birth",
    ):
        super().__init__(label)

    def get_prompt(self):
        _p = get_translator(context=True)
        return _p("country-select", "What country are you from?")


class CountryOfResidence(CountryDropdown):
    def __init__(
        self,
        label="country_of_residence",
    ):
        super().__init__(label)

    def get_prompt(self):
        _p = get_translator(context=True)
        return _p("country-select", "What is your current country of residence?")


class FormalEducation(ModularPage):
    def __init__(
        self,
        label="formal_education",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "formal-education", "What is your highest level of formal education?"
        )
        self.time_estimate = 5

        control = RadioButtonControl(
            [
                "none",
                "high_school",
                "college",
                "graduate_school",
                "postgraduate_degree_or_higher",
            ],
            [
                _p("formal-education", "None"),
                _p("formal-education", "High school"),
                _p("formal-education", "College"),
                _p("formal-education", "Graduate School"),
                _p("formal-education", "Postgraduate degree or higher"),
            ],
            name="formal_education",
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


# Language #
class MotherTongue(ModularPage):
    def __init__(
        self,
        label="mother_tongue",
        # TODO Change back to plural (add "(s)") once multi-select is implemented.
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "language-select",
            "What is your mother tongue - i.e., the language which you have grown up speaking from early childhood?",
        )
        self.time_estimate = 5
        locale = get_locale()
        language_dict = get_language_dict(locale)

        control = DropdownControl(
            choices=list(language_dict.keys()) + ["other"],
            labels=list(language_dict.values()) + ["Other language"],
            default_text=_p("language-select", "Select a language"),
            name=self.label,
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )

    def validate(self, response, **kwargs):
        _p = get_translator(context=True)
        if self.control.force_selection and response.answer == "":
            return FailedValidation(
                _p("language-select", "You need to select a language!")
            )
        return None


class MoreThanOneLanguage(ModularPage):
    def __init__(
        self,
        label="more_than_one_language",
    ):
        _ = get_translator()
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p("language-select", "Do you speak more than one language?")
        self.time_estimate = 5

        control = PushButtonControl(
            choices=["yes", "no"],
            labels=[_("Yes"), _("No")],
            arrange_vertically=False,
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


class LanguagesInOrderOfProficiency(ModularPage):
    def __init__(
        self,
        label="languages_in_order_of_proficiency",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "language-select",
            "Please list the languages you speak in order of proficiency (first language first, second language second, ...)",
        )
        self.time_estimate = 5
        super().__init__(
            self.label,
            self.prompt,
            control=TextControl(),
            time_estimate=self.time_estimate,
            save_answer=label,
        )

    def validate(self, response, **kwargs):
        _p = get_translator(context=True)
        if not response.answer != "":
            return FailedValidation(
                _p("language-select", "Please list at least one language!")
            )
        return None


# Basic music #
class YearsOfFormalTraining(ModularPage):
    def __init__(
        self,
        label="years_of_formal_training",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "music",
            "How many years of formal training on a musical instrument (including voice) have you had during your lifetime?",
        )
        self.time_estimate = 5
        super().__init__(
            self.label,
            self.prompt,
            control=NumberControl(),
            time_estimate=self.time_estimate,
            save_answer=label,
        )


class HoursOfDailyMusicListening(ModularPage):
    def __init__(
        self,
        label="hours_of_daily_music_listening",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "music", "On average, how many hours do you listen to music daily?"
        )
        self.time_estimate = 5
        super().__init__(
            self.label,
            self.prompt,
            control=NumberControl(),
            time_estimate=self.time_estimate,
            save_answer=label,
        )


class MoneyFromPlayingMusic(ModularPage):
    def __init__(
        self,
        label="money_from_playing_music",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p("music", "Do you make money from playing music?")
        self.time_estimate = 5

        control = RadioButtonControl(
            ["frequently", "sometimes", "never"],
            [_p("music", "Frequently"), _p("music", "Sometimes"), _p("music", "Never")],
            name="money_from_playing_music",
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


# Hearing loss #
class HearingLoss(ModularPage):
    def __init__(
        self,
        label="hearing_loss",
    ):
        _ = get_translator()
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "music", "Do you have hearing loss or any other hearing issues?"
        )
        self.time_estimate = 5

        control = PushButtonControl(
            choices=["yes", "no"],
            labels=[_("Yes"), _("No")],
            arrange_vertically=False,
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


# Dance #
class DanceSociallyOrProfessionally(ModularPage):
    def __init__(
        self,
        label="dance_socially_or_professionally",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p("dance", "Do you dance socially or professionally?")
        self.time_estimate = 5

        control = RadioButtonControl(
            ["socially", "professionally", "never_dance"],
            [
                _p("dance", "Socially"),
                _p("dance", "Professionally"),
                _p("dance", "I never dance"),
            ],
            name="dance_socially_or_professionally",
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


class LastTimeDanced(ModularPage):
    def __init__(
        self,
        label="last_time_danced",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "dance",
            "When was the last time you danced? (choose the most accurate answer):",
        )
        self.time_estimate = 5

        control = RadioButtonControl(
            [
                "this_week",
                "this_month",
                "this_year",
                "some_years_ago",
                "many_years_ago",
                "never_danced",
            ],
            [
                _p("dance", "This week"),
                _p("dance", "This month"),
                _p("dance", "This year"),
                _p("dance", "Some years ago"),
                _p("dance", "Many years ago"),
                _p("dance", "I never danced"),
            ],
            name="last_time_danced",
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


# Speech disorders #
class SpeechLanguageTherapy(ModularPage):
    def __init__(
        self,
        label="speech_language_therapy",
    ):
        _ = get_translator()
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "speech-disorder", "Did you get speech-language therapy as a child?"
        )
        self.time_estimate = 5

        control = PushButtonControl(
            choices=["yes", "no", "dont_know"],
            labels=[_("Yes"), _("No"), _("I don’t know")],
            arrange_vertically=False,
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


class DiagnosedWithDyslexia(ModularPage):
    def __init__(
        self,
        label="diagnosed_with_dyslexia",
    ):
        _ = get_translator()
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "speech-disorder", "Have you ever been diagnosed with dyslexia?"
        )
        self.time_estimate = 5

        control = PushButtonControl(
            choices=["yes", "no", "dont_know"],
            labels=[_("Yes"), _("No"), _("I don’t know")],
            arrange_vertically=False,
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


# Income #
class HouseholdIncomePerYear(ModularPage):
    def __init__(
        self,
        label="household_income_per_year",
        currency="USD",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p("income", "What is your total household income per year?")
        self.time_estimate = 5

        control = RadioButtonControl(
            [
                "ĺess_than_10000",
                "10000_to_19999",
                "20000_to_29999",
                "30000_to_39999",
                "40000_to_49999",
                "50000_to_59999",
                "60000_to_69999",
                "70000_to_79999",
                "80000_to_89999",
                "90000_to_99999",
                "100000_to_149999",
                "150000_or_more",
            ],
            [
                _p("income", "Less than 10,000 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "10,000 to 19,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "20,000 to 29,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "30,000 to 39,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "40,000 to 49,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "50,000 to 59,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "60,000 to 69,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "70,000 to 79,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "80,000 to 89,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "90,000 to 99,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "100,000 to 149,999 {CURRENCY}").format(CURRENCY=currency),
                _p("income", "150,000 {CURRENCY} or more").format(CURRENCY=currency),
            ],
            name="household_income_per_year",
        )
        super().__init__(
            self.label,
            self.prompt,
            control=control,
            time_estimate=self.time_estimate,
            save_answer=label,
        )


# ExperimentFeedback #
class LikedExperiment(ModularPage):
    def __init__(
        self,
        label="liked_experiment",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p("experiment-feedback", "Did you like the experiment?")
        self.time_estimate = 5
        super().__init__(
            self.label,
            self.prompt,
            control=TextControl(
                bot_response=lambda: "I'm a bot so I don't really have feelings..."
            ),
            time_estimate=self.time_estimate,
            save_answer=label,
        )


class FoundExperimentDifficult(ModularPage):
    def __init__(
        self,
        label="find_experiment_difficult",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = _p(
            "experiment-feedback", "Did you find the experiment difficult?"
        )
        self.time_estimate = 5
        super().__init__(
            self.label,
            self.prompt,
            control=TextControl(),
            time_estimate=self.time_estimate,
            bot_response=lambda: "I'm a bot so I found it pretty easy...",
            save_answer=label,
        )


class EncounteredTechnicalProblems(ModularPage):
    def __init__(
        self,
        label="encountered_technical_problems",
    ):
        _p = get_translator(context=True)
        self.label = label
        self.prompt = (
            _p(
                "experiment-feedback",
                "Did you encounter any technical problems during the experiment?",
            )
            + " "
            + _p(
                "experiment-feedback",
                "If so, please provide a few words describing the problem.",
            )
        )
        self.time_estimate = 5
        super().__init__(
            self.label,
            self.prompt,
            control=TextControl(),
            time_estimate=self.time_estimate,
            bot_response=lambda: "No technical problems.",
            save_answer=label,
        )



import inspect
from random import sample

from dallinger import db
from dallinger.transformations import Transformation
from sqlalchemy import Column
from sqlalchemy.orm import declared_attr, deferred

from psynet.field import PythonObject
from psynet.trial import ChainNode
from psynet.trial.chain import ChainTrial
from psynet.trial.main import TrialMaker
from psynet.utils import get_logger

logger = get_logger()


def get_super_classes(cls):
    return inspect.getmro(cls)


def get_extended_class(obj):
    super_classes = get_super_classes(obj.__class__)
    mixin_class = super_classes[1]
    extended_class_idx = len(inspect.getmro(mixin_class))
    return super_classes[extended_class_idx]


def assert_correct_inheritance(sub_class, super_class):
    super_classes = get_super_classes(sub_class)
    assert len(super_classes) >= 3, "The class must inherit from at least two classes"
    error_msg = (
        "The mixin must be the first class you inherit from, for example:\n"
        "class CreateAndRateTrialMaker(CreateAndRateTrialMakerMixin, ImitationChainTrialMaker) would be correct and \n"
        "class CreateAndRateTrialMaker(ImitationChainTrialMaker, CreateAndRateTrialMakerMixin) would be incorrect."
    )
    assert super_classes[1] == super_class, error_msg


def filter_relevant_kwargs(kwargs, func):
    params = inspect.getfullargspec(func)
    keys = params.args + params.kwonlyargs
    return {key: kwargs[key] for key in keys if key in kwargs}


def get_required_parameters(func):
    inspected_function = inspect.getfullargspec(func)
    args = inspected_function.args + [
        arg
        for arg in inspected_function.kwonlyargs
        if arg not in inspected_function.kwonlydefaults
    ]
    args.remove("self")

    if inspected_function.defaults is None:
        return args
    else:
        return args[: -len(inspected_function.defaults)]


def split_kwargs(obj, kwargs, trial_maker_class, mixin_class):
    required_parameters = list(
        set(
            get_required_parameters(trial_maker_class.__init__)
            + get_required_parameters(obj.setup)
        )
    )
    for param in required_parameters:
        assert param in kwargs, f"Missing required parameter: {param}"
    trial_maker_kwargs = filter_relevant_kwargs(kwargs, trial_maker_class.__init__)
    mixin_kwargs = filter_relevant_kwargs(kwargs, mixin_class.setup)
    return trial_maker_kwargs, mixin_kwargs


class CreateAndRateTrialMixin(object):
    def __init__(self, experiment, node, participant, *args, **kwargs):
        trial_class = get_extended_class(self)
        assert issubclass(
            trial_class, ChainTrial
        ), "The trial class must inherit from ChainTrial"
        super().__init__(experiment, node, participant, *args, **kwargs)
        trial_class.__init__(self, experiment, node, participant, *args, **kwargs)


class CreateTrialMixin(CreateAndRateTrialMixin):
    pass


class RateOrSelectTrialMixin(CreateAndRateTrialMixin):
    def __init__(self, experiment, node, participant, *args, **kwargs):
        super().__init__(experiment, node, participant, *args, **kwargs)
        self.targets = self.get_targets()
        self.register_transformations(self.targets)

    @declared_attr
    def targets(cls):
        # See the mixin section of https://docs.sqlalchemy.org/en/14/orm/inheritance.html#resolving-column-conflicts
        # return deferred(Column(PythonObject))
        return deferred(cls.__table__.c.get("targets", Column(PythonObject)))

    def register_transformations(self, targets):
        # Register the transformations
        for target in targets:
            if issubclass(target.__class__, ChainTrial):
                transformation = Transformation(info_out=self, info_in=target)
                db.session.add(transformation)
        db.session.commit()

    def get_targets(self):
        return self.get_all_targets()

    def get_all_targets(self):
        trial_maker = self.trial_maker
        shuffle = trial_maker.randomize_target_presentation_order
        assert issubclass(trial_maker.__class__, CreateAndRateTrialMakerMixin)
        creator_class = trial_maker.creator_class
        targets = creator_class.query.filter_by(
            node_id=self.node_id, failed=False, finalized=True
        ).all()
        if self.trial_maker.include_previous_iteration:
            targets += [self.node]
        if shuffle:
            targets = sample(targets, len(targets))
        return targets

    def get_target_answer(self, target):
        if issubclass(target.__class__, ChainNode):
            return target.definition
        elif issubclass(target.__class__, ChainTrial):
            return target.answer
        else:
            raise NotImplementedError()


class SelectTrialMixin(RateOrSelectTrialMixin):
    def get_targets(self):
        assert self.trial_maker.target_selection_method == "all"
        return self.get_all_targets()

    def format_answer(self, answer, **kwargs):
        rated_target_strs = [f"{target}" for target in self.targets]
        assert (
            answer in rated_target_strs
        ), "The answer must be one of the rated target_strs"
        return answer


class RateTrialMixin(RateOrSelectTrialMixin):
    def get_targets(self):
        target_selection_method = self.trial_maker.target_selection_method
        if target_selection_method == "all":
            return self.get_all_targets()
        elif target_selection_method == "one":
            return self.get_one_target()
        else:
            raise NotImplementedError(
                f"Unknown rated_targets value: {target_selection_method}"
            )

    def count_rated_targets(self, available_targets_str):
        rater_class = self.trial_maker.rater_class
        all_rating_trials = rater_class.query.filter_by(
            node_id=self.node_id, failed=False
        ).all()
        all_rating_trials = [
            trial for trial in all_rating_trials if trial.id != self.id
        ]
        all_rated_target_strs = [
            f"{target}" for rating in all_rating_trials for target in rating.targets
        ]
        target2count = dict(
            zip(available_targets_str, [0] * len(available_targets_str))
        )
        for target_str in all_rated_target_strs:
            target2count[target_str] += 1
        return target2count

    def select_target_with_least_ratings(self, all_targets):
        all_targets_str = [f"{target}" for target in all_targets]
        target2count = self.count_rated_targets(all_targets_str)

        min_count = min(target2count.values())
        targets_with_min_count = [
            target_str
            for target_str, count in target2count.items()
            if count == min_count
        ]

        target_str_with_least_ratings = sample(targets_with_min_count, 1)[0]

        if self.trial_maker.verbose:
            logger.info(
                f"For network {self.network.id} at iteration {self.node.degree} we have the following"
                + f" ratings for: {target2count}. We therefore selected: {target_str_with_least_ratings}."
            )

        target_idx = all_targets_str.index(target_str_with_least_ratings)
        return all_targets[target_idx]

    def get_one_target(self):
        return [self.select_target_with_least_ratings(self.get_all_targets())]

    def format_answer(self, answer, **kwargs):
        rated_target_strs = [f"{target}" for target in self.targets]
        if len(self.targets) > 1:
            assert type(answer) is list, "The answer must be a list of ratings"
            assert len(answer) == len(
                self.targets
            ), "The answer must have the same length as the number of targets"
            assert all(
                [type(rating) in [int, float] for rating in answer]
            ), "The answer must be a list of numbers"
            answer = dict(zip(rated_target_strs, answer))
        else:
            if isinstance(answer, str):
                float_answer = float(answer)
                int_answer = int(answer)
                answer = float_answer if float_answer != int_answer else int_answer
            assert type(answer) in [int, float], "The answer must be a number"
            assert len(rated_target_strs) == 1
            answer = {rated_target_strs[0]: answer}
        return answer


class CreateAndRateNodeMixin(object):
    def __init__(self, **kwargs):
        extended_class = get_extended_class(self)
        assert issubclass(
            extended_class, ChainNode
        ), "The extended class must be a ChainNode"
        extended_class.__init__(self, **kwargs)

    def get_str2target(self, rate_or_select_trials):
        all_targets = rate_or_select_trials[0].get_all_targets()
        return {f"{target}": target for target in all_targets}

    def summarize_rate_trials(self, rate_trials):
        import numpy as np

        str2target = self.get_str2target(rate_trials)
        all_target_strs = list(str2target.keys())
        rating_dict = {target_str: [] for target_str in all_target_strs}
        for rate_trial in rate_trials:
            for target_str, rating in rate_trial.answer.items():
                rating_dict[target_str] += [rating]
        mean_rating_dict = {
            target_str: np.mean(ratings) for target_str, ratings in rating_dict.items()
        }
        target_str_with_highest_rating = max(mean_rating_dict, key=mean_rating_dict.get)

        if self.trial_maker.verbose:
            logger.info(
                f"For network {self.network_id} at iteration {self.degree} we have the following"
                + f" ratings for: {mean_rating_dict}. We therefore selected: {target_str_with_highest_rating}."
            )
        return str2target[target_str_with_highest_rating]

    def summarize_select_trials(self, select_trials):
        str2target = self.get_str2target(select_trials)
        count_dict = {target_str: 0 for target_str in str2target.keys()}
        for trial in select_trials:
            count_dict[trial.answer] += 1
        target_str_with_highest_count = max(count_dict, key=count_dict.get)
        if self.trial_maker.verbose:
            logger.info(
                f"For network {self.network_id} at iteration {self.degree} we have the following"
                + f" ratings for: {count_dict}. We therefore selected: {target_str_with_highest_count}."
            )
        return str2target[target_str_with_highest_count]

    def summarize_trials(self, trials, experiment, participant):
        trial_maker = self.trial_maker
        all_rate_trials = trial_maker.rater_class.query.filter_by(
            node_id=self.id, failed=False, finalized=True
        ).all()

        rate_mode = trial_maker.rate_mode

        if rate_mode == "rate":
            return self.summarize_rate_trials(all_rate_trials)
        elif rate_mode == "select":
            return self.summarize_select_trials(all_rate_trials)
        else:
            raise NotImplementedError(f"Unknown rate_mode value: {rate_mode}")


class CreateAndRateNode(CreateAndRateNodeMixin, ChainNode):
    def create_initial_seed(self, experiment, participant):
        return {}

    def create_definition_from_seed(self, seed, experiment, participant):
        return seed


class CreateAndRateTrialMakerMixin(object):
    def __init__(self, **kwargs):
        assert_correct_inheritance(self.__class__, CreateAndRateTrialMakerMixin)
        extended_class = get_extended_class(self)
        assert issubclass(
            extended_class, TrialMaker
        ), "The extended class must be a TrialMaker"
        trial_maker_kwargs, mixin_kwargs = self.prepare_kwargs(
            kwargs, extended_class, CreateAndRateTrialMakerMixin
        )
        self.setup(**mixin_kwargs)
        extended_class.__init__(self, **trial_maker_kwargs)

    def assert_correct_inheritance_elements(self):
        def check_inheritance(clc, base_class, mixin_class):
            assert issubclass(clc, mixin_class)
            assert issubclass(clc, base_class)

        check_inheritance(self.node_class, ChainNode, CreateAndRateNodeMixin)
        check_inheritance(self.creator_class, ChainTrial, CreateTrialMixin)
        if self.rate_mode == "select":
            check_inheritance(self.rater_class, ChainTrial, SelectTrialMixin)
        elif self.rate_mode == "rate":
            check_inheritance(self.rater_class, ChainTrial, RateTrialMixin)
        else:
            raise NotImplementedError(f"Unknown rate_mode value: {self.rate_mode}")

    def setup(
        self,
        n_creators,
        n_raters,
        node_class,
        creator_class,
        rater_class,
        start_nodes,
        rate_mode="rate",
        include_previous_iteration=False,
        target_selection_method="one",
        randomize_target_order=True,
        verbose=False,
    ):
        # Set class variables
        self.n_creators = n_creators
        self.n_raters = n_raters
        self.rate_mode = rate_mode
        self.creator_class = creator_class
        self.rater_class = rater_class
        self.node_class = node_class
        self.include_previous_iteration = include_previous_iteration
        self.n_rate_stimuli = self.n_creators + int(self.include_previous_iteration)
        self.target_selection_method = target_selection_method
        self.randomize_target_presentation_order = randomize_target_order
        self.verbose = verbose

        # Assertions
        self.assert_is_positive_integer(n_creators)
        self.assert_is_positive_integer(n_raters)
        self.assert_correct_inheritance_elements()

        if self.include_previous_iteration:
            error_msg = (
                "If you want to include previous iterations, you need to specify the seed, e.g."
                " CreateAndRateNode(seed='My initial response shown to the participant')"
            )
            n_nodes_with_seed = sum(
                [node.seed is not None and len(node.seed) > 0 for node in start_nodes]
            )
            assert len(start_nodes) == n_nodes_with_seed, error_msg

        if self.rate_mode == "rate":
            if self.target_selection_method == "one":
                if self.include_previous_iteration:
                    error_msg = (
                        "n_raters must be a multiple of n_creators + 1 (since include_previous_iteration == "
                        "True) if rate_mode is 'rate' and target_selection_method is 'one'."
                    )
                else:
                    error_msg = (
                        "n_raters must be a multiple of n_creators if rate_mode is 'rate' and "
                        "target_selection_method is 'one'."
                    )
                assert self.n_raters % (self.n_rate_stimuli) == 0, error_msg
                self.n_rate_stimuli = 1
        elif self.rate_mode == "select":
            assert (
                self.n_rate_stimuli > 1
            ), '`n_rate_stimuli` must be greater than 1 if `rate_mode` is "select"'
        else:
            raise NotImplementedError(f"Unknown rate_mode value: {rate_mode}")

        assert target_selection_method in ["one", "all"]

    @staticmethod
    def assert_is_positive_integer(x):
        assert type(x) is int and x > 0, f"{x} must be a positive integer"

    def prepare_kwargs(self, kwargs, trial_maker_class, mixin_class):
        kwargs["trial_class"] = kwargs["creator_class"]
        kwargs["node_class"] = kwargs["node_class"]
        trial_maker_kwargs, mixin_kwargs = split_kwargs(
            self, kwargs, trial_maker_class, mixin_class
        )
        trials_per_node = kwargs["n_creators"] + kwargs["n_raters"]
        trial_maker_kwargs["trials_per_node"] = trials_per_node
        return trial_maker_kwargs, mixin_kwargs

    def get_trial_class(self, node, participant, experiment):
        create_trials = self.get_non_failed_creations(node)
        finished_creations = self.get_finished_creations(node)
        need_creators = len(create_trials) < self.n_creators
        waiting_for_creators = len(finished_creations) < len(create_trials)

        if need_creators:
            return self.creator_class
        else:
            if waiting_for_creators:
                return None
            else:
                return self.rater_class

    def get_non_failed_creations(self, node):
        return self.creator_class.query.filter_by(node_id=node.id, failed=False).all()

    def get_finished_creations(self, node):
        return self.creator_class.query.filter_by(
            node_id=node.id, failed=False, finalized=True
        ).all()



from typing import Optional

from psynet.modular_page import ModularPage, PushButtonControl
from psynet.page import InfoPage
from psynet.timeline import Module, join


class PEI(Module):
    """
    Confidence scale (PEI) questionnaire.

    Parameters
    ----------

    label : str, default: "pei"
        A label used to distinguish the module from other modules in the timeline.

    info_page : InfoPage, optional, default: `None`
        An :class:`~psynet.page.InfoPage` object to be used as an introductionary first page.
        If none is supplied the default one is displayed (see source code).
    """

    def __init__(
        self,
        label: str = "pei",
        info_page: Optional[InfoPage] = None,
    ):
        if info_page is None:
            info_page = InfoPage(
                "The following questions are designed to assess your confidence. Please answer as accurately as possible. Think about how the question applies to you during the last 2 months.",
                time_estimate=5,
            )
        self.label = label
        self.elts = join(
            info_page,
            PEI_01(),
            PEI_02(),
            PEI_03(),
            PEI_04(),
            PEI_05(),
            PEI_06(),
            PEI_07(),
        )
        super().__init__(self.label, self.elts)


class PEI_01(ModularPage):
    def __init__(
        self,
        label="pei_01",
        prompt="I often feel unsure of myself even in situations I have successfully dealt with in the past.",
    ):
        self.label = label
        self.prompt = prompt
        self.time_estimate = 5

        control = PushButtonControl(
            agreement_scale()["choices"],
            agreement_scale()["labels"],
        )
        super().__init__(
            self.label, self.prompt, control=control, time_estimate=self.time_estimate
        )


class PEI_02(ModularPage):
    def __init__(
        self,
        label="pei_02",
        prompt="I lack some important capabilities that may keep me from being successful.",
    ):
        self.label = label
        self.prompt = prompt
        self.time_estimate = 5

        control = PushButtonControl(
            agreement_scale()["choices"],
            agreement_scale()["labels"],
        )
        super().__init__(
            self.label, self.prompt, control=control, time_estimate=self.time_estimate
        )


class PEI_03(ModularPage):
    def __init__(
        self,
        label="pei_03",
        prompt="Much of the time I don’t feel as competent as many of the people around me.",
    ):
        self.label = label
        self.prompt = prompt
        self.time_estimate = 5

        control = PushButtonControl(
            agreement_scale()["choices"],
            agreement_scale()["labels"],
        )
        super().__init__(
            self.label, self.prompt, control=control, time_estimate=self.time_estimate
        )


class PEI_04(ModularPage):
    def __init__(
        self,
        label="pei_04",
        prompt="I have fewer doubts about my abilities than most people.",
    ):
        self.label = label
        self.prompt = prompt
        self.time_estimate = 5

        control = PushButtonControl(
            agreement_scale()["choices"],
            agreement_scale()["labels"],
        )
        super().__init__(
            self.label, self.prompt, control=control, time_estimate=self.time_estimate
        )


class PEI_05(ModularPage):
    def __init__(
        self,
        label="pei_05",
        prompt="When things are going poorly, I am usually confident that I can successfully deal with them.",
    ):
        self.label = label
        self.prompt = prompt
        self.time_estimate = 5

        control = PushButtonControl(
            agreement_scale()["choices"],
            agreement_scale()["labels"],
        )
        super().__init__(
            self.label, self.prompt, control=control, time_estimate=self.time_estimate
        )


class PEI_06(ModularPage):
    def __init__(
        self,
        label="pei_06",
        prompt="I have more confidence in myself than most people I know.",
    ):
        self.label = label
        self.prompt = prompt
        self.time_estimate = 5

        control = PushButtonControl(
            agreement_scale()["choices"],
            agreement_scale()["labels"],
        )
        super().__init__(
            self.label, self.prompt, control=control, time_estimate=self.time_estimate
        )


class PEI_07(ModularPage):
    def __init__(
        self,
        label="pei_07",
        prompt="If I were more confident about myself, my life would be better.",
    ):
        self.label = label
        self.prompt = prompt
        self.time_estimate = 5

        control = PushButtonControl(
            agreement_scale()["choices"],
            agreement_scale()["labels"],
        )
        super().__init__(
            self.label, self.prompt, control=control, time_estimate=self.time_estimate
        )


def agreement_scale():
    return {
        "choices": list(range(1, 5)),
        "labels": [
            "Strongly Agree",
            "Mainly Agree",
            "Mainly Disagree",
            "Strongly Disagree",
        ],
    }



import json
import os
import re
from datetime import datetime, timedelta
from math import ceil

import dallinger.recruiters
import dominate
import flask
import pandas as pd
import requests
import sqlalchemy
from dallinger import db
from dallinger.config import get_config
from dallinger.db import session
from dallinger.notifications import admin_notifier, get_mailer
from dallinger.recruiters import RedisStore
from dallinger.utils import get_base_url
from dominate import tags
from dominate.util import raw
from sqlalchemy import Column, DateTime, Float, ForeignKey, Integer, String
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.sql import func

from .consent import AudiovisualConsent, LucidConsent, OpenScienceConsent
from .data import SQLBase, SQLMixin, register_table
from .lucid import get_lucid_service
from .participant import Participant
from .timeline import Response, TimelineLogic
from .utils import get_logger, get_translator, render_template_with_translations

logger = get_logger()


class PsyNetRecruiterMixin:
    show_termination_button = False

    def terminate_participant(
        self, participant=None, assignment_id=None, reason=None, details=None
    ):
        raise NotImplementedError

    def release_participant(self, experiment, participant) -> TimelineLogic:
        from .page import ExecuteFrontEndJS

        return ExecuteFrontEndJS(
            "dallinger.submitAssignment()",
            message="Communicating with the recruiter...",
        )

    def check_consents(self, consents):
        """
        Check that the consent elements are suitable for the recruiter.
        By default this check is skipped in ``psynet debug local``.

        Parameters
        ----------
        consents : list
            List of consent objects from the timeline
        """
        if len(consents) == 0:
            raise RuntimeError(
                "It looks like your experiment is missing a consent page. "
                "Is that right? You can resolve this check by adding a pre-prepared consent page from psynet.consent "
                "to your timeline, or a custom subclass of psynet.consent.Consent, "
                "or psynet.consent.NoConsent to skip this check entirely."
            )


class HotAirRecruiter(PsyNetRecruiterMixin, dallinger.recruiters.HotAirRecruiter):
    pass


class ProlificRecruiter(PsyNetRecruiterMixin, dallinger.recruiters.ProlificRecruiter):
    pass


class DevProlificRecruiter(
    PsyNetRecruiterMixin, dallinger.recruiters.DevProlificRecruiter
):
    pass


class MTurkRecruiter(PsyNetRecruiterMixin, dallinger.recruiters.MTurkRecruiter):
    pass


# CAP Recruiter
class BaseCapRecruiter(PsyNetRecruiterMixin, dallinger.recruiters.CLIRecruiter):
    """
    The CapRecruiter base class
    """

    def recruit(self, n=1):
        """Incremental recruitment isn't implemented for now, so we return an empty list."""
        return []

    def open_recruitment(self, n=1):
        """
        Return an empty list which otherwise would be a list of recruitment URLs.
        """
        return {"items": [], "message": ""}

    def close_recruitment(self):
        logger.info("No more participants required. Recruitment stopped.")

    def notify_duration_exceeded(self, participants, reference_time):
        """
        The participant has been working longer than the time defined in
        the "duration" config value.
        """
        for participant in participants:
            participant.status = "abandoned"
            # We preserve this commit just in case Dallinger removes the external commit in the future
            session.commit()

    def reward_bonus(self, participant, amount, reason):
        """
        Return values for `basePay` and `bonus` to cap-recruiter application.
        """
        data = {
            "assignmentId": participant.assignment_id,
            "basePayment": self.config.get("base_payment"),
            "bonus": amount,
            "failed_reason": participant.failure_tags,
        }
        url = self.external_submission_url
        url += "/fail" if participant.failed else "/complete"

        requests.post(
            url,
            json=data,
            headers={"Authorization": os.environ.get("CAP_RECRUITER_AUTH_TOKEN")},
            verify=False,  # Temporary fix because of SSLCertVerificationError
        )


class CapRecruiter(BaseCapRecruiter):
    """
    The production cap-recruiter.

    """

    nickname = "cap-recruiter"
    external_submission_url = "https://cap-recruiter.ae.mpg.de/tasks"


class StagingCapRecruiter(BaseCapRecruiter):
    """
    The staging cap-recruiter.

    """

    nickname = "staging-cap-recruiter"
    external_submission_url = "https://staging-cap-recruiter.ae.mpg.de/tasks"


class DevCapRecruiter(BaseCapRecruiter):
    """
    The development cap-recruiter.

    """

    nickname = "dev-cap-recruiter"
    external_submission_url = "http://localhost:8000/tasks"


# Lucid Recruiter
@register_table
class LucidRID(SQLBase, SQLMixin):
    __tablename__ = "lucid_rid"

    # These fields are removed from the database table as they are not needed.
    failed = None
    failed_reason = None
    time_of_death = None
    vars = None
    creation_time = None

    rid = Column(String, ForeignKey("participant.worker_id"), index=True)
    registered_at = Column(DateTime, server_default=func.now())
    updated_at = Column(DateTime, onupdate=func.now())

    completed_at = Column(DateTime)
    terminated_at = Column(DateTime)
    termination_reason = Column(String)
    termination_details = Column(String)

    # Lucid fields
    lucid_status = Column(String)
    lucid_status_code = Column(Integer)
    lucid_fulcrum_status = Column(Integer)
    lucid_market_place_code = Column(String)
    lucid_entry_date = Column(DateTime)
    lucid_last_date = Column(DateTime)
    lucid_panelist_id = Column(String)
    lucid_respondent_id = Column(String)
    lucid_supplier_id = Column(Integer)

    # to dict
    def to_dict(self):
        return {
            "rid": self.rid,
            "registered_at": self.registered_at,
            "updated_at": self.updated_at,
            "completed_at": self.completed_at,
            "terminated_at": self.terminated_at,
            "termination_reason": self.termination_reason,
            "termination_details": self.termination_details,
            "lucid_status": self.lucid_status,
            "lucid_status_code": self.lucid_status_code,
            "lucid_fulcrum_status": self.lucid_fulcrum_status,
            "lucid_market_place_code": self.lucid_market_place_code,
            "lucid_entry_date": self.lucid_entry_date,
            "lucid_last_date": self.lucid_last_date,
            "lucid_panelist_id": self.lucid_panelist_id,
            "lucid_respondent_id": self.lucid_respondent_id,
            "lucid_supplier_id": self.lucid_supplier_id,
        }


@register_table
class LucidStatus(SQLBase, SQLMixin):
    __tablename__ = "lucid_status"

    # These fields are removed from the database table as they are not needed.
    failed = None
    failed_reason = None
    time_of_death = None
    vars = None

    status = Column(String)
    cost = Column(Float)
    currency = Column(String)
    exchange_rate = Column(Float)
    cost_per_survey = Column(Float)
    payment_per_hour = Column(Float)
    earnings_per_click = Column(Float)
    system_conversion = Column(Integer)
    completion_loi = Column(Integer)
    termination_loi = Column(Integer)
    last_complete_date = Column(DateTime)

    total_entrants = Column(Integer)
    total_completes = Column(Integer)
    total_screens = Column(Integer)
    drop_off_rate = Column(Float)
    conversion_rate = Column(Float)
    incidence_rate = Column(Float)

    def to_dict(self):
        return {
            "timestamp": self.creation_time,
            "status": self.status,
            "cost": self.cost,
            "currency": self.currency,
            "exchange_rate": self.exchange_rate,
            "cost_per_survey": self.cost_per_survey,
            "payment_per_hour": self.payment_per_hour,
            "earnings_per_click": self.earnings_per_click,
            "system_conversion": self.system_conversion,
            "completion_loi": self.completion_loi,
            "termination_loi": self.termination_loi,
            "last_complete_date": self.last_complete_date,
            "total_entrants": self.total_entrants,
            "total_screens": self.total_screens,
            "total_completes": self.total_completes,
            "drop_off_rate": self.drop_off_rate,
            "conversion_rate": self.conversion_rate,
            "incidence_rate": self.incidence_rate,
        }


class LucidRecruiterException(Exception):
    """Custom exception for LucidRecruiter"""


class BaseLucidRecruiter(PsyNetRecruiterMixin, dallinger.recruiters.CLIRecruiter):
    MARKETPLACE_CODE = "Marketplace codes"
    IN_SURVEY = "Currently in Client Survey or Drop"
    COMPLETED = "Returned as Complete"
    TERMINATED = "Returned as Terminate"
    SURVEY_CLOSED = "Survey Closed"
    survey_codes = ["awarded", "pending", "paused", "live", "complete", "archived"]
    client_codes = {
        # See https://support.lucidhq.com/s/article/Client-Response-Codes
        -1: MARKETPLACE_CODE,
        1: IN_SURVEY,
        10: COMPLETED,  # Returned as Complete from PsyNet
        11: COMPLETED,  # Adjusted Complete
        20: TERMINATED,  # Terminated from PsyNet
        26: TERMINATED,  # Adjusted Terminate
        28: TERMINATED,  # Adjusted Terminate
        30: TERMINATED,  # Quality termination
        33: TERMINATED,  # Speeder
        34: TERMINATED,  # Open End Terminate
        35: TERMINATED,  # Encryption Failure
        38: TERMINATED,  # Adjusted to Terminate
        134: TERMINATED,  # Encryption Failure at Client Survey
        135: TERMINATED,  # Encryption Failure at Marketplace Return
        136: TERMINATED,  # Survey Closed
        137: TERMINATED,  # Verify Callback Failure
        233: TERMINATED,  # Invalid Client Response Status
        235: TERMINATED,  # Secure Client Callback Failure
        40: TERMINATED,  # Client Survey Quota Full
        60: TERMINATED,  # Quality Terminate on Pre-Client Intermediary Page
        62: TERMINATED,  # Declined Routing on Pre-Client Intermediary Page
        66: TERMINATED,  # Declined Routing on Pre-Client Intermediary Page
        91: TERMINATED,  # Incorrectly Formatted Redirect
        110: TERMINATED,  # Used for specific opt-in studies
        70: COMPLETED,  # Audience: Returned as Complete
        80: TERMINATED,  # Audience: Returned as Terminate
    }

    market_place_codes = {
        -6: "Sent to Marketplace Intermediate",
        -5: "Sent to External Intermediate",
        -1: "Error",
        1: "In Screener",
        3: "In Client Survey",
        21: "Industry Lockout",
        23: "Standard Qualification",
        24: "Custom Qualification",
        120: "Pre-Client Survey Opt Out",
        122: "Return to Marketplace Opt Out",
        123: "Max Client Survey Entries",
        124: "Max Time in Router",
        125: "Max Time in Router Warning Opt Out",
        126: "Max Answer Limit",
        30: "Quality Term: Unique IP",
        31: "Quality Term: RelevantID Duplicate",
        32: "Quality Term: Invalid Traffic",
        35: "Quality Term: Supplier PID Duplicate",
        36: "Quality Term: Cookie Duplicate",
        37: "Quality Term: GEO IP Mismatch",
        38: "Quality Term: RelevantID** Fraud Profile",
        131: "Quality Term: Supplier Encryption Failure",
        132: "Quality Term: Blocked PID",
        133: "Quality Term: Blocked IP",
        134: "Quality Term: Max Completes per Day Terminate",
        138: "Quality Term: Survey Group Cookie Duplicate",
        139: "Quality Term: Survey Group Supplier PID Duplicate",
        230: "Quality Term: Survey Group Unique IP",
        234: "OFAC Term: Blocked Country IP",
        236: "Privacy Term: No Privacy Consent",
        237: "Privacy Term: Minimum Age",
        238: "Quality Term: Found on Deny List",
        240: "Quality Term: Invalid Browser",
        241: "Quality Term: Respondent Threshold Limit",
        242: "Quality Term: Respondent Quality Score",
        243: "Quality Term: Marketplace Signature Check",
        40: "Overquota: Quota Full",
        41: "Overquota: Supplier Allocation",
        42: "Overquota: Survey Closed for Entry",
        50: "Financial Term: CPI Below Supplier’s Rate Card",
        98: "Exit: End of Router",
    }

    """
    The LucidRecruiter base class
    """

    show_termination_button = True

    required_consent_page = LucidConsent.LucidConsentPage
    optional_consent_pages = (
        AudiovisualConsent.AudiovisualConsentPage,
        OpenScienceConsent.OpenScienceConsentPage,
    )

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.config = get_config()
        if self.config.get("show_reward"):
            raise RuntimeError(
                "Lucid recruitment requires `show_reward` to be set to `False`."
            )
        self.mailer = get_mailer(self.config)
        self.notifies_admin = admin_notifier(self.config)
        recruitment_config = json.loads(self.config.get("lucid_recruitment_config"))

        self.lucidservice = get_lucid_service(self.config, recruitment_config)
        self.store = kwargs.get("store", RedisStore())

    def recruit(self, n=1):
        """Incremental recruitment isn't implemented for now, so we return an empty list."""
        return []

    def notify_duration_exceeded(self, participants, reference_time):
        """
        The participant has been working longer than the time defined in
        the "duration" config value.
        """
        for participant in participants:
            participant.status = "abandoned"
            # We preserve this commit just in case Dallinger removes the external commit in the future
            session.commit()

    def run_checks(self):
        logger.info("Polling Lucid API to count entry_df")
        survey_number = self.current_survey_number()
        respondents = pd.DataFrame(self.lucidservice.get_submissions(survey_number))
        summary = self.lucidservice.get_summary(survey_number)
        total_completes = summary["total_completes"]
        logger.info(
            f"Found {summary['total_entrants']} entrants, {summary['total_screens']} after_screener, {total_completes} completes"
        )

        cost = summary["cost"]
        currency = summary["currency"]
        completion_loi = summary["completion_loi"]
        cost_per_survey = (cost / total_completes) if total_completes > 0 else 0
        payment_per_hour = completion_loi / 60 * cost_per_survey
        drop_off_rate = 0
        conversion_rate = 0
        incidence_rate = 0

        if len(respondents) > 0:
            respondents["status"] = respondents.client_status.apply(
                lambda x: self.client_codes.get(x, "Unknown")
            )
            respondents["market_place_code"] = respondents.fulcrum_status.apply(
                lambda x: self.market_place_codes.get(x, "Unknown")
            )

            all_entrants = LucidRID.query.all()
            entrants_dict = {entrant.rid: entrant for entrant in all_entrants}

            lucid_entrants = []

            for _, row in respondents.iterrows():
                if row.respondent_id in entrants_dict:
                    entrant = entrants_dict[row.respondent_id]
                    changed = False
                    fields_to_update = {
                        "lucid_status": "status",
                        "lucid_status_code": "client_status",
                        "lucid_fulcrum_status": "fulcrum_status",
                        "lucid_market_place_code": "market_place_code",
                        "lucid_last_date": "last_date",
                    }
                    for field, api_field in fields_to_update.items():
                        if getattr(entrant, field) != row[api_field]:
                            setattr(entrant, field, row[api_field])
                            changed = True
                    if changed:
                        db.session.add(entrant)
                else:
                    entrant = LucidRID(
                        rid=row.respondent_id,
                        lucid_status=row.status,
                        lucid_status_code=row.client_status,
                        lucid_fulcrum_status=row.fulcrum_status,
                        lucid_market_place_code=row.market_place_code,
                        lucid_entry_date=row.entry_date,
                        lucid_last_date=row.last_date,
                        lucid_panelist_id=row.panelist_id,
                        lucid_respondent_id=row.respondent_id,
                        lucid_supplier_id=row.supplier_id,
                    )
                    db.session.add(entrant)
                lucid_entrants.append(entrant)

            entry_df = pd.DataFrame([entrant.to_dict() for entrant in lucid_entrants])
            MARKETPLACE_CODE = self.MARKETPLACE_CODE  # noqa: F841
            COMPLETED_CODE = self.COMPLETED  # noqa: F841
            IN_SURVEY_CODE = self.IN_SURVEY  # noqa: F841
            after_screener = entry_df.query("lucid_status != @MARKETPLACE_CODE")
            completes = entry_df.query("lucid_status == @COMPLETED_CODE")
            in_survey = entry_df.query("lucid_status == @IN_SURVEY_CODE")
            drop_off_rate = (
                len(in_survey) / len(after_screener) if len(after_screener) > 0 else 0
            )
            conversion_rate = (
                len(completes) / len(after_screener) if len(after_screener) > 0 else 0
            )

            pattern = "Privacy Term|Quality Term|Financial Term|OFAC Term|Custom Qualification|Standard Qualification"
            n_returned_because_of_qualifications = (
                entry_df.lucid_market_place_code.str.contains(pattern, regex=True).sum()
            )

            n_potential_completes = (
                len(completes) + n_returned_because_of_qualifications
            )
            incidence_rate = float(
                len(completes) / n_potential_completes
                if n_potential_completes > 0
                else 0.0
            )

        logger.info(f"Payment per hour: {payment_per_hour:.2f} {currency}")
        logger.info(f"Drop off rate: {drop_off_rate:.2%}")
        logger.info(f"Conversion rate: {conversion_rate:.2%}")
        logger.info(f"Incidence rate: {conversion_rate:.2%}")

        status_entry = LucidStatus(
            # From the summary
            status=summary["status"],
            cost=cost,
            currency=currency,
            exchange_rate=summary["exchange_rate"],
            cost_per_survey=cost_per_survey,
            payment_per_hour=payment_per_hour,
            earnings_per_click=summary["epc"],
            system_conversion=summary["system_conversion"],
            completion_loi=completion_loi,
            termination_loi=summary["termination_loi"],
            last_complete_date=summary["last_complete_date"],
            # From the metrics
            total_entrants=summary["total_entrants"],
            total_completes=summary["total_completes"],
            total_screens=summary["total_screens"],
            drop_off_rate=drop_off_rate,
            conversion_rate=conversion_rate,
            incidence_rate=incidence_rate,
        )
        db.session.add(status_entry)
        db.session.commit()

        unfailed_entrants = LucidRID.query.filter_by(
            terminated_at=None, completed_at=None
        ).all()
        logger.info(f"Found {len(unfailed_entrants)} of which are not failed")
        now = datetime.now()

        for entrant in unfailed_entrants:
            if (
                entrant.registered_at
                + timedelta(seconds=self.initial_response_within_s)
                > now
            ):
                # skip entrants that have not been registered long enough
                continue
            if entrant.completed_at is not None:
                # skip completed entrants
                continue
            if entrant.terminated_at is not None:
                # skip terminated entrants
                continue

            details = None
            participant = None
            reason = None
            try:
                participant = Participant.query.filter_by(worker_id=entrant.rid).one()
                responses = (
                    Response.query.filter_by(participant_id=participant.id)
                    .order_by(Response.creation_time)
                    .all()
                )
                if len(responses) == 0:
                    reason = "first-response-timeout"

            except sqlalchemy.orm.exc.NoResultFound:
                # Do not terminate participants who did not pass the qualifications
                if entrant.lucid_status != self.MARKETPLACE_CODE:
                    reason = "never-entered-experiment"

            if reason:
                try:
                    participant_info = (
                        {"participant": participant}
                        if participant
                        else {"assignment_id": entrant.rid}
                    )
                    self.terminate_participant(
                        reason=reason, details=details, **participant_info
                    )

                    logger.info(
                        f"Successfully terminated participant with RID '{entrant.rid}'."
                    )
                except Exception as e:
                    logger.error(
                        f"Error terminating participant with RID '{entrant.rid}': {e}"
                    )

    def get_survey_storage_key(self, name):
        experiment_id = self.config.get("id")
        return f"{self.__class__.__name__}:{experiment_id}:{name}"

    @property
    def in_progress(self):
        """Does a Lucid survey for the current experiment ID already exist?"""
        return self.current_survey_number() is not None

    def check_consents(self, consents):
        super().check_consents(consents)
        error_msg = "Lucid recruitment requires consent 'LucidConsent' and optionally one of `AudiovisualConsent` or `OpenScienceConsent` (in this order)."
        if isinstance(consents[0], self.required_consent_page):
            if len(consents) == 1:
                pass
            elif len(consents) == 2 and isinstance(
                consents[1], self.optional_consent_pages
            ):
                pass
            else:
                raise RuntimeError(error_msg)
        else:
            raise RuntimeError(error_msg)

    def current_survey_number(self):
        """
        Return the survey number associated with the active experiment ID
        if any such survey exists.
        """
        return self.store.get(self.get_survey_storage_key("survey_number"))

    def current_survey_sid(self):
        """
        Return the survey SID associated with the active experiment ID
        if any such survey exists.
        """
        return self.store.get(self.get_survey_storage_key("survey_sid"))

    def open_recruitment(self, n=1):
        """Open a connection to Lucid and create a survey."""
        from .experiment import get_experiment
        from .utils import get_config

        self.lucidservice.log(f"Opening initial recruitment for {n} participants.")
        if self.in_progress:
            raise LucidRecruiterException(
                "Tried to open recruitment on already open recruiter."
            )

        experiment = get_experiment()
        wage_per_hour = get_config().get("wage_per_hour")
        estimated_duration = experiment.estimated_completion_time(wage_per_hour)
        create_survey_request_params = {
            "bid_length_of_interview": ceil(estimated_duration / 60),
            "live_url": self.ad_url.replace("http://", "https://"),
            "name": self.config.get("title"),
            "quota": n,
            "quota_cpi": round(
                experiment.estimated_max_reward(wage_per_hour),
                2,
            ),
        }

        survey_info = self.lucidservice.create_survey(
            self.config.get("publish_experiment"), **create_survey_request_params
        )
        self._record_current_survey_number(survey_info["SurveyNumber"])
        self._record_survey_sid(survey_info["SurveySID"])

        # Lucid Marketplace automatically adds 6 qualifications to US studies
        # when a survey is created (Age, Gender, Zip, Ethnicity, Hispanic, Standard HHI US).
        # We update the qualifications in this case to remove these constraints on the participants.
        # See https://developer.lucidhq.com/#post-create-a-survey
        survey_number = self.current_survey_number()
        if self.lucidservice.recruitment_config["survey"]["CountryLanguageID"] == 9:
            self.lucidservice.remove_default_qualifications_from_survey(survey_number)

        self.lucidservice.add_qualifications_to_survey(survey_number)

        url = survey_info["ClientSurveyLiveURL"]
        self.lucidservice.log(
            f"Done creating Lucid project and survey: {survey_number}."
        )
        self.lucidservice.log(
            f"Lucid reports: https://marketplace.samplicio.us/fulcrum/next/surveys/{survey_number}/reports"
        )
        self.lucidservice.log("---------> " + url)
        self.lucidservice.log("----------")

        survey_id = self.current_survey_number()
        if survey_id is None:
            self.lucidservice.log("No survey in progress: Recruitment aborted.")
            return

        lucid_url = (
            f"https://marketplace.samplicio.us/fulcrum/next/surveys/{survey_id}/quotas"
        )
        message = f"Lucid survey {survey_id} created successfully. " f"URL: {lucid_url}"

        return {
            "items": [url],
            "message": message,
        }

    def close_recruitment(self):
        """
        Lucid automatically ends recruitment when the number of completes has reached the
        target.
        """
        self.lucidservice.log("Recruitment is automatically handled by Lucid.")

    def normalize_entry_information(self, entry_information):
        """Accepts data from the recruited user and returns data needed to validate,
        create or load a Dallinger Participant.

        See :func:`~dallinger.experiment.Experiment.create_participant` for
        details.

        The default implementation extracts ``hit_id``, ``assignment_id``, and
        ``worker_id`` values directly from ``entry_information``.

        This implementation extracts the ``RID`` from ``entry_information``
        and assigns the value to ``hit_id``, ``assignment_id``, and ``worker_id``.
        """

        rid = entry_information.get("RID")
        hit_id = entry_information.get("hit_id")
        if hit_id is None:
            hit_id = entry_information.get("hitId")

        if rid is None and hit_id is None:
            raise LucidRecruiterException(
                "Either `RID` or `hit_id` has to be present in `entry_information`."
            )

        if rid is None:
            rid = hit_id

        # Save RID info into the database
        try:
            LucidRID.query.filter_by(rid=rid).one()
        except NoResultFound:
            self.lucidservice.log(f"Saving RID '{rid}' into the database.")
            db.session.add(LucidRID(rid=rid))
            db.session.commit()
        except MultipleResultsFound:
            raise MultipleResultsFound(
                f"Multiple rows for Lucid RID '{rid}' found. This should never happen."
            )

        participant_data = {
            "hit_id": rid,
            "assignment_id": rid,
            "worker_id": rid,
        }

        if entry_information:
            participant_data["entry_information"] = entry_information

        return participant_data

    def exit_response(self, experiment, participant):
        """
        Delegate to the experiment for possible values to show to the
        participant and complete the survey.
        """
        external_submit_url = self.external_submit_url(participant=participant)
        self.lucidservice.log(f"Exit redirect: {external_submit_url}")

        return render_template_with_translations(
            "exit_recruiter_lucid.html",
            external_submit_url=external_submit_url,
        )

    def reward_bonus(self, participant, amount, reason):
        """
        Set `completed_at` timestamp on participant's LucidRID entry
        """
        if participant is not None and participant.progress == 1:
            self.complete_participant(participant.assignment_id)
        else:
            responses = (
                Response.query.filter_by(participant_id=participant.id)
                .order_by(Response.creation_time)
                .all()
            )
            if responses[-1].answer == {"lucid_consent": False}:
                reason = "consent-rejected"
            else:
                reason = "participant-did-not-complete"
            self.terminate_participant(participant=participant, reason=reason)

    def _record_current_survey_number(self, survey_number):
        self.store.set(self.get_survey_storage_key("survey_number"), survey_number)

    def _record_survey_sid(self, survey_sid):
        self.store.set(self.get_survey_storage_key("survey_sid"), survey_sid)

    def external_submit_url(self, participant=None, assignment_id=None):
        if participant is None and assignment_id is None:
            raise RuntimeError(
                "Error generating 'external_submit_url': One of 'participant' or 'assignment_id' needs to be provided."
            )
        data = self.data_for_submit_url(participant, assignment_id)
        return self.lucidservice.generate_submit_url(ris=data["ris"], rid=data["rid"])

    def data_for_submit_url(self, participant, assignment_id):
        # Standard terminate
        ris = 20
        if participant is not None:
            assignment_id = participant.assignment_id
            if "performance_check" in participant.failure_tags:
                # Security terminate
                ris = 30
            elif participant.progress == 1:
                # Complete
                ris = 10
        if assignment_id is None:
            assignment_id = assignment_id
        return {"rid": assignment_id, "ris": ris}

    def error_page_content(self, assignment_id, external_submit_url):
        _p = get_translator(context=True)

        if external_submit_url is None:
            external_submit_url = self.external_submit_url(assignment_id=assignment_id)

        html = tags.div()
        with html:
            tags.p(
                " ".join(
                    [
                        _p(
                            "lucid_error",
                            "Redirecting to Lucid Marketplace...",
                        ),
                    ]
                )
            )
            tags.script(
                raw(
                    'setTimeout(() => { window.location = "'
                    + external_submit_url
                    + '"; }, 2000)'
                )
            )
        return html

    def time_until_termination_in_s(self, rid):
        return self.lucidservice.time_until_termination_in_s(rid)

    def complete_participant(self, rid):
        return self.lucidservice.complete_respondent(rid)

    def terminate_participant(
        self, participant=None, assignment_id=None, reason=None, details=None
    ):
        assert participant or assignment_id
        assert not (participant and assignment_id)

        if participant:
            assignment_id = participant.assignment_id

            participant.failed = True
            participant.failed_reason = reason
            participant.status = "returned"
            db.session.commit()
        try:
            self.lucidservice.terminate_respondent(assignment_id, reason, details)
            logger.info(
                f"Terminating respondent with RID '{assignment_id}'. Reason: '{reason}'"
            )
        except Exception as e:
            logger.error(
                f"Error terminating respondent with RID '{assignment_id}': {e}"
            )

        return self.external_submit_url(assignment_id=assignment_id)

    def set_termination_details(self, rid, reason):
        self.lucidservice.set_termination_details(rid, reason)

    def get_config_entry(self, key):
        lucid_recruitment_config = json.loads(
            self.config.get("lucid_recruitment_config")
        )

        return lucid_recruitment_config.get(key)

    def get_participant(self, request):
        assignment_id = request.values.get("assignmentId")
        unique_id = request.values.get("unique_id")
        participant_id = request.values.get("participant_id")
        rid = request.values.get("RID")
        participant = None

        if assignment_id is None:
            if unique_id is not None:
                assignment_id = unique_id.split(":")[1]
            elif rid is not None:
                assignment_id = rid
            elif participant_id is not None:
                participant = (
                    Participant.query.with_for_update(of=Participant)
                    .populate_existing()
                    .get(int(participant_id))
                )
                assignment_id = participant.assignment_id

        assert assignment_id is not None, "Could not determine assignment_id."

        if participant is None:
            try:
                participant = Participant.query.filter_by(
                    assignment_id=assignment_id
                ).one()
            except NoResultFound:
                logger.error(
                    f"No LucidRID for Lucid RID '{assignment_id}' found. This should never happen."
                )
            except MultipleResultsFound:
                logger.error(
                    f"Multiple rows for Lucid RID '{assignment_id}' found. This should never happen."
                )

        return participant

    @property
    def termination_time_in_s(self):
        return self.get_config_entry("termination_time_in_s")

    @property
    def inactivity_timeout_in_s(self):
        return self.get_config_entry("inactivity_timeout_in_s")

    @property
    def no_focus_timeout_in_s(self):
        return self.get_config_entry("no_focus_timeout_in_s")

    @property
    def aggressive_no_focus_timeout_in_s(self):
        return self.get_config_entry("aggressive_no_focus_timeout_in_s")

    @property
    def initial_response_within_s(self):
        return self.get_config_entry("initial_response_within_s")

    def get_status(self):
        query = LucidStatus.query.order_by(LucidStatus.id.desc())
        recruiter_info = super().get_status()
        if query.count() > 0:
            recruiter_info = {**recruiter_info, **query.first().to_dict()}
            recruiter_info["total_working"] = LucidRID.query.filter_by(
                terminated_at=None, completed_at=None
            ).count()
        return recruiter_info

    def change_lucid_status(self, status):
        survey_number = self.current_survey_number()
        service = get_lucid_service()
        service.change_status(survey_number, status)
        LucidStatus.query.order_by(LucidStatus.id.desc()).first().status = status
        db.session.commit()


class DevLucidRecruiter(BaseLucidRecruiter):
    """
    Development recruiter for the Lucid Marketplace.
    """

    nickname = "dev-lucid-recruiter"

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.ad_url = (
            f"http://localhost.cap:5000/ad?recruiter={self.nickname}&RID=[%RID%]"
        )


class LucidRecruiter(BaseLucidRecruiter):
    """
    The production Lucid recruiter.
    Recruit participants from the Lucid Marketplace.
    """

    nickname = "lucid-recruiter"

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.ad_url = f"{get_base_url()}/ad?recruiter={self.nickname}&RID=[%RID%]"


def get_lucid_country_language_id(country_tag, language_tag, service=None):
    assert len(country_tag) == 2, "Country tag must be 2 characters long."
    assert country_tag.isupper(), "Country tag must be uppercase."
    assert len(language_tag) == 3, "Language tag must be 3 characters long."
    assert language_tag.isupper(), "Language tag must be uppercase."

    if service is None:
        service = get_lucid_service()
    lookup = service.get_lucid_country_language_lookup()
    selection = lookup.query(
        "country_tag == @country_tag and language_tag == @language_tag"
    )
    if len(selection) == 0:
        pd.set_option("display.max_rows", None)
        raise ValueError(
            f"Could not find country language ID for {country_tag} and {language_tag}. Pick from these:\n{lookup}"
        )
    return selection.iloc[0]["Id"]


def get_lucid_settings(
    lucid_recruitment_config_path,
    termination_time_in_s: int,
    bid_incidence=66,
    collects_pii=False,
    inactivity_timeout_in_s=120,
    no_focus_timeout_in_s=60,
    aggressive_no_focus_timeout_in_s=3,
    initial_response_within_s=180,
    debug_recruiter=False,
):
    """
    Parameters
    ----------
    lucid_recruitment_config_path: str, path to the Lucid recruitment config.

    termination_time_in_s: int, maximal time a participant can spend on the experiment. If this time is exceeded,
        the participant is terminated via the front-end.

    bid_incidence: int, default 66, the bid incidence. Bid incidence is the number of completes/(number of completes +
        participants who did not pass the qualifications). It is a percentage, so if you expect 66% of the participants
        to pass the qualifications, set it to 66. Set it to a realistic value, but as high as possible.

    collects_pii: bool, default False, whether the survey collects personally identifiable information.

    inactivity_timeout_in_s: int, default 120, the inactivity timeout in seconds. If the participant is inactive for
        this amount of time, the participant is terminated via the front-end. Inactive means that the participant does
        not interact with the page (i.e., no ["click", "keypress", "load", "mousedown", "mousemove", "touchstart"]).

    no_focus_timeout_in_s: int, default 60, the no focus timeout in seconds. If the participant moves the mouse outside
        the window or opens another tab, the participant is terminated via the front-end after this amount of time.

    aggressive_termination_on_no_focus: int, default 3, this the same setting as `no_focus_timeout_in_s`, but it is
        only used for aggressive in the consent page, since many participants are lost there.

    initial_response_within_s: int, default 180 seconds (3 minutes). If the participant does not proceed to the consent
        within this time, the participant is terminated via the backend-end.

    debug_recruiter: bool, default False, whether to use the development recruiter. This is useful for local testing.

    """

    with open(lucid_recruitment_config_path, "r") as f:
        lucid_recruitment_config = json.load(f)

    if termination_time_in_s is not None:
        lucid_recruitment_config["termination_time_in_s"] = termination_time_in_s

    lucid_recruitment_config["survey"]["BidIncidence"] = bid_incidence
    lucid_recruitment_config["survey"]["CollectsPII"] = collects_pii
    lucid_recruitment_config["inactivity_timeout_in_s"] = inactivity_timeout_in_s
    lucid_recruitment_config["no_focus_timeout_in_s"] = no_focus_timeout_in_s
    lucid_recruitment_config["aggressive_no_focus_timeout_in_s"] = (
        aggressive_no_focus_timeout_in_s
    )
    lucid_recruitment_config["initial_response_within_s"] = initial_response_within_s

    lucid_recruitment_config = json.dumps(lucid_recruitment_config)

    settings = {
        "recruiter": "LucidRecruiter",
        "lucid_recruitment_config": lucid_recruitment_config,
        "currency": "EUR",
        "show_reward": False,
        "show_abort_button": False,
    }
    if debug_recruiter:
        settings["debug_recruiter"] = "DevLucidRecruiter"
    return settings


class GenericRecruiter(PsyNetRecruiterMixin, dallinger.recruiters.CLIRecruiter):
    """
    An improved version of Dallinger's Hot-Air Recruiter.
    """

    nickname = "generic"

    def recruit(self, n=1):
        return []

    def exit_response(self, experiment, participant):
        from psynet.timeline import Page

        message = experiment.render_exit_message(participant)

        if message is None:
            raise ValueError(
                "experiment.render_exit_message returned None. Did you forget to use 'return'?"
            )

        elif isinstance(message, Page):
            raise ValueError(
                "Sorry, you can't return a Page from experiment.render_exit_message."
            )

        elif message == "default_exit_message":
            return super().exit_response(experiment, participant)

        elif isinstance(message, str):
            html = dominate.tags.p(message).render()

        elif isinstance(message, dominate.dom_tag.dom_tag):
            html = message.render()

        else:
            raise ValueError(
                f"Invalid value of experiment.render_exit_message: {message}. "
                "You should return either a string or an HTML specification created using dominate tags "
                "(see https://pypi.org/project/dominate/)."
            )

        return flask.render_template("custom_html.html", html=html)

    def open_recruitment(self, n=1):
        res = super().open_recruitment(n=n)

        # Hide the Dallinger logs advice, because the advice doesn't work for SSH deployment
        res["message"] = re.sub(
            "Open the logs for this experiment.*", "", res["message"]
        )
        res["message"] = re.sub(
            ".*in the logs for subsequent recruitment URLs\\.", "", res["message"]
        )

        return res

    def notify_duration_exceeded(self, participants, reference_time):
        """
        The participant has been working longer than the time defined in
        the "duration" config value.
        """
        for participant in participants:
            participant.status = "abandoned"
            # We preserve this commit just in case Dallinger removes the external commit in the future
            session.commit()



import datetime

import pandas as pd
from flask import render_template


def report_resource_use():
    TEMPLATE_NAME = "dashboard_resources.html"
    title = "Resource usage"
    data = summarize_resource_use()
    if len(data) == 0:
        return render_template(
            TEMPLATE_NAME,
            title=title,
            html="""
            <div class="alert alert-danger" role="alert">
                Wait at least 1 minute to see the first data.
            </div>
            """,
        )

    return render_template(
        TEMPLATE_NAME,
        title=title,
        html="",
        data=data,
    )


def summarize_resource_use():
    from psynet.experiment import ExperimentStatus

    window_length = datetime.timedelta(hours=48)
    data = (
        ExperimentStatus.query.filter(
            ExperimentStatus.creation_time > datetime.datetime.now() - window_length
        )
        .order_by(ExperimentStatus.id.desc())
        .all()
    )

    if len(data) == 0:
        return None

    df_raw = pd.DataFrame([row.to_dict() for row in data])
    df_raw["free_disk_space"] = df_raw["free_disk_space_gb"]
    df_raw.drop(columns=["extra_info", "id", "free_disk_space_gb"], inplace=True)

    df_normalized = normalize_resource_use(df_raw)

    df_plot = df_normalized.melt(id_vars="timestamp", var_name="type", value_name="y")
    df_plot = add_raw_values(df_plot, df_raw)
    df_plot["label"] = df_plot.apply(format_label, axis=1)
    df_plot["timestamp"] = format_time_str(df_plot["timestamp"])
    df_plot["type"] = format_unit(df_plot["type"])

    return df_plot.to_dict(orient="records")


def format_label(row):
    match row.type:
        case "cpu_usage_pct":
            return f"{row.y_unit} % of total CPU usage"
        case "ram_usage_pct":
            return f"{row.y_unit} % of total RAM"
        case "free_disk_space":
            return f"{int(row.y_unit)} GB free disk space"
        case "median_response_time":
            return f"{round(row.y_unit, 2)} s median response time"
        case "requests_per_minute":
            return f"{int(row.y_unit)} page loads per minute"
        case "n_working_participants":
            return f"{int(row.y_unit)} total working participants"
        case _:
            return row.y_unit


def max_100(x):
    return (x / x.max()) * 100


def normalize_resource_use(_resources_df):
    resources_df = _resources_df.copy()
    resources_df["timestamp"] = resources_df.index
    resources_df["free_disk_space"] = 100 - max_100(resources_df["free_disk_space"])
    resources_df["median_response_time"] = max_100(resources_df["median_response_time"])
    resources_df["requests_per_minute"] = max_100(resources_df["requests_per_minute"])
    resources_df["n_working_participants"] = max_100(
        resources_df["n_working_participants"]
    )
    return resources_df


def add_raw_values(df_plot, df_raw):
    df_raw_long = df_raw.melt(id_vars="timestamp", var_name="type", value_name="y")
    df_plot["y_unit"] = df_raw_long["y"]
    df_plot["x"] = df_plot["timestamp"].astype(int)
    df_plot["timestamp"] = df_raw_long["timestamp"]
    df_plot.dropna(inplace=True)
    return df_plot


def format_time_str(timestamp_series):
    now = pd.to_datetime("now")
    earliest = timestamp_series.min()

    if now.day == earliest.day:
        date_format = "%H:%M"
    elif now.year == earliest.year:
        date_format = "%m-%d %H:%M"
    else:
        date_format = "%Y-%m-%d %H:%M"

    return [
        str(ts)
        for ts in pd.to_datetime(timestamp_series, unit="s").dt.strftime(date_format)
    ]


def format_unit(type_list: list):
    replacement_dict = {
        "cpu_usage_pct": "CPU usage (%)",
        "ram_usage_pct": "RAM usage (%)",
        "free_disk_space": "Used disk space compared to min (%)",
        "median_response_time": "Median page loading time (%)",
        "requests_per_minute": "Number of page loads",
        "n_working_participants": "Total working participants",
    }
    return [replacement_dict.get(item, item) for item in type_list]



# pylint: disable=attribute-defined-outside-init

import json
from smtplib import SMTPAuthenticationError
from typing import TYPE_CHECKING, Dict

import dallinger.models
from dallinger import db
from dallinger.notifications import admin_notifier
from sqlalchemy import (
    Boolean,
    Column,
    Float,
    ForeignKey,
    Integer,
    String,
    UniqueConstraint,
    desc,
)
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship
from sqlalchemy.orm.collections import attribute_mapped_collection

from .asset import AssetParticipant
from .data import SQLMixinDallinger
from .field import PythonList, PythonObject, VarStore, extra_var
from .utils import call_function_with_context, get_config, get_logger, organize_by_key

logger = get_logger()

if TYPE_CHECKING:
    from .sync import SyncGroup
    from .timeline import Module

# pylint: disable=unused-import

UniqueConstraint(dallinger.models.Participant.worker_id)
UniqueConstraint(dallinger.models.Participant.unique_id)


class Participant(SQLMixinDallinger, dallinger.models.Participant):
    """
    Represents an individual participant taking the experiment.
    The object is linked to the database - when you make changes to the
    object, it should be mirrored in the database.

    Users should not have to instantiate these objects directly.

    The class extends the ``Participant`` class from base Dallinger
    (:class:`dallinger.models.Participant`) to add some useful features,
    in particular the ability to store arbitrary variables.

    The following attributes are recommended for external use:

    * :attr:`~psynet.participant.Participant.answer`
    * :attr:`~psynet.participant.Participant.var`
    * :attr:`~psynet.participant.Participant.failure_tags`

    The following method is recommended for external use:

    * :meth:`~psynet.participant.Participant.append_failure_tags`

    See below for more details.

    Attributes
    ----------

    id : int
        The participant's unique ID.

    elt_id : list
        Represents the participant's position in the timeline.
        Should not be modified directly.
        The position is represented as a list, where the first element corresponds
        to the index of the participant within the timeline's underlying
        list representation, and successive elements (if any) represent
        the participant's position within (potentially nested) page makers.
        For example, ``[10, 3, 2]`` would mean go to
        element 10 in the timeline (0-indexing),
        which must be a page maker;
        go to element 3 within that page maker, which must also be a page maker;
        go to element 2 within that page maker.

    elt_bounds : list
        Represents the number of elements at each level of the current
        ``elt_id`` hierarchy; used to work out when to leave a page maker
        and go up to the next level.
        Should not be modified directly.

    page_uuid : str
        A long unique string that is randomly generated when the participant advances
        to a new page, used as a passphrase to guarantee the security of
        data transmission from front-end to back-end.
        Should not be modified directly.

    complete : bool
        Whether the participant has successfully completed the experiment.
        A participant is considered to have successfully completed the experiment
        once they hit a :class:`~psynet.timeline.SuccessfulEndPage`.
        Should not be modified directly.

    aborted : bool
        Whether the participant has aborted the experiment.
        A participant is considered to have aborted the experiment
        once they have hit the "Abort experiment" button on the "Abort experiment" confirmation page.

    answer : object
        The most recent answer submitted by the participant.
        Can take any form that can be automatically serialized to JSON.
        Should not be modified directly.

    response : Response
        An object of class :class:`~psynet.timeline.Response`
        providing detailed information about the last response submitted
        by the participant. This is a more detailed version of ``answer``.

    branch_log : list
        Stores the conditional branches that the participant has taken
        through the experiment.
        Should not be modified directly.

    failure_tags : list
        Stores tags that identify the reason that the participant has failed
        the experiment (if any). For example, if a participant fails
        a microphone pre-screening test, one might add "failed_mic_test"
        to this tag list.
        Should be modified using the method :meth:`~psynet.participant.Participant.append_failure_tags`.

    var : :class:`~psynet.field.VarStore`
        A repository for arbitrary variables; see :class:`~psynet.field.VarStore` for details.

    progress : float [0 <= x <= 1]
        The participant's estimated progress through the experiment.

    client_ip_address : str
        The participant's IP address as reported by Flask.

    answer_is_fresh : bool
        ``True`` if the current value of ``participant.answer`` (and similarly ``participant.last_response_id`` and
        ``participant.last_response``) comes from the last page that the participant saw, ``False`` otherwise.

    browser_platform : str
        Information about the participant's browser version and OS platform.

    all_trials : list
        A list of all trials for that participant.

    alive_trials : list
        A list of all non-failed trials for that participant.

    failed_trials : list
        A list of all failed trials for that participant.
    """

    # We set the polymorphic_identity manually to differentiate the class
    # from the Dallinger Participant class.
    __extra_vars__ = {}

    elt_id = Column(PythonList)
    elt_id_max = Column(PythonList)

    time_credit = Column(Float)
    estimated_max_time_credit = Column(Float)
    progress = Column(Float)

    # If time_credit_fixes is non-empty, then the last element of time_credit_fixes
    # is used as the currently active time credit 'fix'. While this fix is active,
    # the participant's time credit will not be allowed to increase above this number.
    # Once the fix expires, the participant's time credit will be set to that number exactly.
    # See ``psynet.timeline.with_fixed_time_credit`` for an explanation.
    time_credit_fixes = Column(PythonList)
    # progress_fixes is analogous to time_credit_fixes, but for progress.
    progress_fixes = Column(PythonList)

    page_uuid = Column(String)
    aborted = Column(Boolean, default=False)
    complete = Column(Boolean, default=False)
    answer = Column(PythonObject)
    answer_accumulators = Column(PythonList)
    sequences = Column(PythonList)
    branch_log = Column(PythonObject)
    for_loops = Column(PythonObject, default=lambda: {})
    failure_tags = Column(PythonList, default=lambda: [])

    base_payment = Column(Float)
    performance_reward = Column(Float)
    unpaid_bonus = Column(Float)
    total_wait_page_time = Column(Float)
    client_ip_address = Column(String, default=lambda: "")
    answer_is_fresh = Column(Boolean, default=False)
    browser_platform = Column(String, default="")
    module_state_id = Column(Integer, ForeignKey("module_state.id"))
    module_state = relationship(
        "ModuleState", foreign_keys=[module_state_id], post_update=True, lazy="selectin"
    )
    current_trial_id = Column(Integer, ForeignKey("info.id"))
    current_trial = relationship(
        "psynet.trial.main.Trial", foreign_keys=[current_trial_id], lazy="joined"
    )
    trial_status = Column(String)

    all_responses = relationship("psynet.timeline.Response")

    # @property
    # def current_trial(self):
    #     if self.in_module and hasattr(self.module_state, "current_trial"):
    #         return self.module_state.current_trial
    #
    # @current_trial.setter
    # def current_trial(self, value):
    #     self.module_state.current_trial = value

    @property
    def last_response(self):
        return self.response

    # all_trials = relationship("psynet.trial.main.Trial")

    @property
    def alive_trials(self):
        return [t for t in self.all_trials if not t.failed]

    @property
    def failed_trials(self):
        return [t for t in self.all_trials if t.failed]

    @property
    def trials(self):
        raise RuntimeError(
            "The .trials attribute has been removed, please use .all_trials, .alive_trials, or .failed_trials instead."
        )

    # This would be better, but we end up with a circular import problem
    # if we try and read csv files using this foreign key...
    #
    # last_response = relationship(
    #     "psynet.timeline.Response", foreign_keys=[last_response_id]
    # )

    # current_trial_id = Column(
    #     Integer, ForeignKey("info.id")
    # )  # 'info.id' because trials are stored in the info table

    # This should work but it's buggy, don't know why.
    # current_trial = relationship(
    #     "psynet.trial.main.Trial",
    #     foreign_keys="[psynet.participant.Participant.current_trial_id]",
    # )
    #
    # Instead we resort to the below...

    # @property
    # def current_trial(self):
    #     from dallinger.models import Info
    #
    #     # from .trial.main import Trial
    #
    #     if self.current_trial_id is None:
    #         return None
    #     else:
    #         # We should just be able to use Trial for the query, but using Info seems
    #         # to avoid an annoying SQLAlchemy bug that comes when we run multiple demos
    #         # in one session. When this happens, what we see is that Trial.query.all()
    #         # sees all trials appropriately, but Trial.query.filter_by(id=1).all() fails.
    #         #
    #         # return Trial.query.filter_by(id=self.current_trial_id).one()
    #         return Info.query.filter_by(id=self.current_trial_id).one()
    #
    # @current_trial.setter
    # def current_trial(self, trial):
    #     from psynet.trial.main import Trial
    #     self.current_trial_id = trial.id if isinstance(trial, Trial) else None

    asset_links = relationship(
        "AssetParticipant",
        collection_class=attribute_mapped_collection("local_key"),
        cascade="all, delete-orphan",
    )

    assets = association_proxy(
        "asset_links",
        "asset",
        creator=lambda k, v: AssetParticipant(local_key=k, asset=v),
    )

    # sync_group_links and sync_groups are defined in sync.py
    # because of import-order necessities

    # sync_groups is a relationship that gives a list of all SyncGroups for that participnat

    @property
    def active_sync_groups(self) -> Dict[str, "SyncGroup"]:
        return {group.group_type: group for group in self.sync_groups if group.active}

    @property
    def sync_group(self) -> "SyncGroup":
        candidates = self.active_sync_groups
        if len(candidates) == 1:
            return list(candidates.values())[0]
        elif len(candidates) == 0:
            return None
        elif len(candidates) > 1:
            raise RuntimeError(
                f"Participant {self.id} is in more than one SyncGroup: "
                f"{list(self.active_sync_groups)}. "
                "Use participant.active_sync_groups[group_type] to access the SyncGroup you need."
            )

    active_barriers = relationship(
        "ParticipantLinkBarrier",
        collection_class=attribute_mapped_collection("barrier_id"),
        cascade="all, delete-orphan",
        primaryjoin=(
            "and_(psynet.participant.Participant.id==remote(ParticipantLinkBarrier.participant_id), "
            "ParticipantLinkBarrier.released==False)"
        ),
        lazy="selectin",
    )

    errors = relationship("ErrorRecord")
    # _module_states = relationship("ModuleState", foreign_keys=[dallinger.models.Participant.id], lazy="selectin")

    @property
    def module_states(self):
        return organize_by_key(
            self._module_states,
            key=lambda x: x.module_id,
            sort_key=lambda x: x.time_started,
        )

    def select_module(self, module_id: str):
        candidates = [
            state
            for state in self._module_states
            if not state.finished and state.module_id == module_id
        ]
        assert len(candidates) == 1
        self.module_state = candidates[0]

    @property
    def var(self):
        return self.globals

    @property
    def globals(self):
        return VarStore(self)

    @property
    def locals(self):
        return self.module_state.var

    def to_dict(self):
        x = SQLMixinDallinger.to_dict(self)
        x.update(self.locals_to_dict())
        return x

    def locals_to_dict(self):
        output = {}
        for module_id, module_states in self.module_states.items():
            module_states.sort(key=lambda x: x.time_started)
            for i, module_state in enumerate(module_states):
                if i == 0:
                    prefix = f"{module_id}__"
                else:
                    prefix = f"{module_id}__{i}__"
                for key, value in module_state.var.items():
                    output[prefix + key] = value
        return output

    @property
    @extra_var(__extra_vars__)
    def aborted_modules(self):
        return [
            log.module_id
            for log in sorted(self._module_states, key=lambda x: x.time_started)
            if log.aborted
        ]

    @property
    @extra_var(__extra_vars__)
    def started_modules(self):
        return [
            log.module_id
            for log in sorted(self._module_states, key=lambda x: x.time_started)
            if log.started
        ]

    @property
    @extra_var(__extra_vars__)
    def finished_modules(self):
        return [
            log.module_id
            for log in sorted(self._module_states, key=lambda x: x.time_started)
            if log.finished
        ]

    def start_module(self, module: "Module"):
        self.check_module_not_already_started(module)
        state = module.state_class(module, self)
        state.start()
        self.module_state = state

    def check_module_not_already_started(self, module: "Module"):
        if module.id not in self.module_states:
            return
        else:
            states = self.module_states[module.id]
            for state in states:
                if not state.finished:
                    raise RuntimeError(
                        f"Participant already has an unfinished module state for '{module.id}'..."
                    )

    def end_module(self, module):
        # This should only fail (delivering multiple logs) if the experimenter has perversely
        # defined a recursive module (or is reusing module ID)
        state = [
            _state for _state in self.module_states[module.id] if not _state.finished
        ]

        if len(state) == 0:
            raise RuntimeError(
                f"Participant had no unfinished module states with id = '{module.id}'."
            )
        elif len(state) > 1:
            raise RuntimeError(
                (
                    f"Participant had multiple unfinished module states with id = '{module.id}': "
                    f"{[s.__json__() for s in state]}, participant: {self.__json__()}"
                )
            )

        state = state[0]
        state.finish()
        self.refresh_module_state()

    def refresh_module_state(self):
        if len(self._module_states) == 0:
            self.module_state = None
        else:
            unfinished = [x for x in self._module_states if not x.finished]
            unfinished.sort(key=lambda x: x.time_started)
            if len(unfinished) == 0:
                self.module_state = None
            else:
                self.module_state = unfinished[-1]

    @property
    def in_module(self):
        return self.module_state is not None

    @property
    @extra_var(__extra_vars__)
    def module_id(self):
        if self.module_state:
            return self.module_state.module_id

    def set_answer(self, value):
        self.answer = value
        return self

    def __init__(self, experiment, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vars = {}
        self.time_credit = 0.0
        self.estimated_max_time_credit = (
            experiment.timeline.estimated_time_credit.get_max("time")
        )
        self.progress = 0.0
        self.time_credit_fixes = []
        self.progress_fixes = []
        self.elt_id = [-1]
        self.elt_id_max = [len(experiment.timeline) - 1]
        self.answer_accumulators = []
        self.sequences = []
        self.complete = False
        self.performance_reward = 0.0
        self.unpaid_bonus = 0.0
        self.base_payment = experiment.base_payment
        self.client_ip_address = None
        self.branch_log = []
        self.total_wait_page_time = 0.0

        db.session.add(self)

        self.initialize(
            experiment
        )  # Hook for custom subclasses to provide further initialization

    def initialize(self, experiment):
        pass

    @property
    def locale(self):
        return self.var.get("locale", default=None)

    @property
    def failure_cascade(self):
        return [lambda: self.alive_trials]

    @property
    def gettext(self):
        return self.translator[0]

    @property
    def pgettext(self):
        return self.translator[1]

    @property
    def time_reward(self):
        wage_per_hour = get_config().get("wage_per_hour")
        seconds = self.time_credit
        hours = seconds / 3600
        return hours * wage_per_hour

    def calculate_reward(self):
        """
        Calculates and returns the currently accumulated reward for the given participant.

        :returns:
            The reward as a ``float``.
        """
        return round(
            self.time_reward + self.performance_reward,
            ndigits=2,
        )

    def inc_time_credit(self, time_credit: float):
        new_value = self.time_credit + time_credit
        new_value = min([new_value, *self.time_credit_fixes])
        self.time_credit = new_value

    def inc_progress(self, time_credit: float):
        if self.estimated_max_time_credit == 0.0:
            new_value = 1.0
        else:
            new_value = self.progress + time_credit / self.estimated_max_time_credit
            new_value = min([new_value, *self.progress_fixes])
        self.progress = new_value

    def inc_performance_reward(self, value):
        self.performance_reward += value

    def amount_paid(self):
        return (0.0 if self.base_payment is None else self.base_payment) + (
            0.0 if self.bonus is None else self.bonus
        )

    def send_email_max_payment_reached(
        self, experiment_class, requested_reward, reduced_reward
    ):
        config = get_config()
        template = """Dear experimenter,

            This is an automated email from PsyNet. You are receiving this email because
            the total amount paid to the participant with assignment_id '{assignment_id}'
            has reached the maximum of {max_participant_payment}$. The reward paid was {reduced_reward}$
            instead of a requested reward of {requested_reward}$.

            The application id is: {app_id}

            To see the logs, use the command "dallinger logs --app {app_id}"
            To pause the app, use the command "dallinger hibernate --app {app_id}"
            To destroy the app, use the command "dallinger destroy --app {app_id}"

            The PsyNet developers.
            """
        message = {
            "subject": "Maximum experiment payment reached.",
            "body": template.format(
                assignment_id=self.assignment_id,
                max_participant_payment=experiment_class.var.max_participant_payment,
                requested_reward=requested_reward,
                reduced_reward=reduced_reward,
                app_id=config.get("id"),
            ),
        }
        logger.info(
            f"Recruitment ended. Maximum amount paid to participant "
            f"with assignment_id '{self.assignment_id}' reached!"
        )
        try:
            admin_notifier(config).send(**message)
        except SMTPAuthenticationError as e:
            logger.error(
                f"SMTPAuthenticationError sending 'max_participant_payment' reached email: {e}"
            )
        except Exception as e:
            logger.error(
                f"Unknown error sending 'max_participant_payment' reached email: {e}"
            )

    @property
    def response(self):
        from .timeline import Response

        return (
            Response.query.filter_by(participant_id=self.id)
            .order_by(desc(Response.id))
            .first()
        )

    def append_branch_log(self, entry: str):
        # We need to create a new list otherwise the change may not be recognized
        # by SQLAlchemy(?)
        if (
            not isinstance(entry, list)
            or len(entry) != 2
            or not isinstance(entry[0], str)
        ):
            raise ValueError(
                f"Log entry must be a list of length 2 where the first element is a string (received {entry})."
            )
        if json.loads(json.dumps(entry)) != entry:
            raise ValueError(
                f"The provided log entry cannot be accurately serialised to JSON (received {entry}). "
                + "Please simplify the log entry (this is typically determined by the output type of the user-provided function "
                + "in switch() or conditional())."
            )
        self.branch_log = self.branch_log + [entry]

    def append_failure_tags(self, *tags):
        """
        Appends tags to the participant's list of failure tags.
        Duplicate tags are ignored.
        See :attr:`~psynet.participant.Participant.failure_tags` for details.

        Parameters
        ----------

        *tags
            Tags to append.

        Returns
        -------

        :class:`psynet.participant.Participant`
            The updated ``Participant`` object.

        """
        original = self.failure_tags
        new = [*tags]
        combined = list(set(original + new))
        self.failure_tags = combined
        return self

    def abort_info(self):
        """
            Information that will be shown to a participant if they click the abort button,
            e.g. in the case of an error where the participant is unable to finish the experiment.

        :returns: ``dict`` which may be rendered to the worker as an HTML table
            when they abort the experiment.
        """
        return {
            "assignment_id": self.assignment_id,
            "hit_id": self.hit_id,
            "accumulated_reward": "$" + "{:.2f}".format(self.calculate_reward()),
        }

    def fail(self, reason=None):
        if self.failed:
            logger.info("Participant %i already failed, not failing again.", self.id)
            return

        if reason is not None:
            self.append_failure_tags(reason)
        reason = ", ".join(self.failure_tags)

        logger.info(
            "Failing participant %i (reason: %s)",
            self.id,
            reason,
        )

        from psynet.experiment import get_experiment

        exp = get_experiment()

        for i, routine in enumerate(exp.participant_fail_routines):
            logger.info(
                "Executing fail routine %i/%i ('%s')...",
                i + 1,
                len(exp.participant_fail_routines),
                routine.label,
            )
            call_function_with_context(
                routine.function,
                participant=self,
                experiment=self,
            )

        super().fail(reason=reason)
        for group in self.active_sync_groups.values():
            from .sync import SimpleSyncGroup

            if isinstance(group, SimpleSyncGroup):
                group.check_numbers()


def get_participant(participant_id: int, for_update: bool = False) -> Participant:
    """
    Returns the participant with a given ID.
    Warning: we recommend just using SQLAlchemy directly instead of using this function.
    When doing so, use ``with_for_update().populate_existing()`` if you plan to update
    this Participant object, that way the database row will be locked appropriately.

    Parameters
    ----------

    participant_id
        ID of the participant to get.

    for_update
        Set to ``True`` if you plan to update this Participant object.
        The Participant object will be locked for update in the database
        and only released at the end of the transaction.

    Returns
    -------

    :class:`psynet.participant.Participant`
        The requested participant.
    """
    query = Participant.query.filter_by(id=participant_id)
    if for_update:
        query = query.with_for_update(of=Participant).populate_existing()
    return query.one()



import datetime
import threading
import time

import dallinger.db
from dallinger import db
from dallinger.db import redis_conn
from rq import Queue
from rq.job import Job
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Float,
    ForeignKey,
    Integer,
    String,
    event,
)
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import deferred, relationship
from tenacity import retry, retry_if_exception_type, stop_after_delay, wait_exponential

from .data import SQLBase, SQLMixin, register_table
from .db import with_transaction
from .field import PythonDict, PythonObject
from .serialize import prepare_function_for_serialization
from .utils import call_function, classproperty, get_logger

logger = get_logger()


@register_table
class AsyncProcess(SQLBase, SQLMixin):
    __tablename__ = "process"
    __extra_vars__ = SQLMixin.__extra_vars__.copy()

    label = Column(String)
    function = Column(PythonObject)
    arguments = deferred(Column(PythonDict))
    pending = Column(Boolean)
    finished = Column(Boolean, default=False)
    time_started = Column(DateTime)
    time_finished = Column(DateTime)
    time_taken = Column(Float)
    _unique_key = Column(PythonDict, unique=True)

    participant_id = Column(Integer, ForeignKey("participant.id"), index=True)
    participant = relationship(
        "psynet.participant.Participant", backref="async_processes"
    )

    trial_maker_id = Column(String, index=True)

    network_id = Column(Integer, ForeignKey("network.id"), index=True)
    network = relationship("TrialNetwork", back_populates="async_processes")

    node_id = Column(Integer, ForeignKey("node.id"), index=True)
    node = relationship("TrialNode", back_populates="async_processes")

    trial_id = Column(Integer, ForeignKey("info.id"), index=True)
    trial = relationship("psynet.trial.main.Trial", back_populates="async_processes")

    response_id = Column(Integer, ForeignKey("response.id"), index=True)
    response = relationship(
        "psynet.timeline.Response", back_populates="async_processes"
    )

    asset_id = Column(Integer, ForeignKey("asset.id"), index=True)
    asset = relationship("Asset", back_populates="async_processes")

    errors = relationship("ErrorRecord")

    launch_queue = []

    def add_to_launch_queue(self):
        self.launch_queue.append(self.get_launch_spec())

    def get_launch_spec(self) -> dict:
        db.session.flush([self])
        return {
            "obj": self,
            "class": self.__class__,
            "id": self.id,
        }

    @classmethod
    def launch_all(cls):
        while cls.launch_queue:
            process = cls.launch_queue.pop(0)
            assert process["obj"].id is not None
            logger.info("Launching async process %s...", process["id"])
            process["class"].launch(process)

    def __init__(
        self,
        function,
        arguments=None,
        trial=None,
        response=None,
        participant=None,
        node=None,
        network=None,
        asset=None,
        label=None,
        unique=False,
    ):
        if label is None:
            label = function.__name__

        if arguments is None:
            arguments = {}

        db.session.flush()

        function, arguments = prepare_function_for_serialization(function, arguments)

        self.label = label
        self.function = function
        self.arguments = arguments

        self.asset = asset
        if asset:
            self.asset_id = asset.id

        self.participant = participant
        if participant:
            self.participant_id = participant.id

        self.network = network
        if network:
            self.network_id = network.id

        self.node = node
        if node:
            self.node_id = node.id

        self.trial = trial
        if trial:
            self.trial_id = trial.id

        self.response = response
        if response:
            self.response_id = response.id

        self.infer_participant()
        self.infer_trial_maker_id()
        self.pending = True

        if unique:
            if isinstance(unique, bool):
                self._unique_key = {
                    "label": label,
                    "function": function,
                    "arguments": arguments,
                }
            else:
                self._unique_key = unique

        db.session.add(self)
        self.add_to_launch_queue()

    def log_time_started(self):
        self.time_started = datetime.datetime.now()

    def log_time_finished(self):
        self.time_finished = datetime.datetime.now()

    def infer_participant(self):
        if self.participant is None:
            for obj in [self.asset, self.trial, self.node, self.network]:
                if obj and hasattr(obj, "participant") and obj.participant:
                    self.participant = obj.participant
                    break
            # For safety...
            if self.participant:
                self.participant_id = self.participant.id

    def infer_trial_maker_id(self):
        for obj in [self.trial, self.node, self.network]:
            if obj and obj.trial_maker_id:
                self.trial_maker_id = obj.trial_maker_id
                return

    @property
    def failure_cascade(self):
        """
        These are the objects that will be failed if the process fails. Ultimately we might want to
        add more objects to this list, for example participants, assets, and networks,
        but currently we're not confident that PsyNet supports failing those objects in that kind of way.
        """
        candidates = [self.trial, self.node]
        return [lambda obj=obj: [obj] for obj in candidates if obj is not None]

    @classmethod
    def launch(cls, process: dict):
        raise NotImplementedError

    @classproperty
    def redis_queue(cls):
        return Queue("default", connection=redis_conn)

    @classmethod
    def call_function_with_logger(cls, process_id):
        cls.call_function(process_id)

    @classmethod
    @retry(
        retry=retry_if_exception_type(NoResultFound),
        wait=wait_exponential(multiplier=0.1, min=0.01),
        stop=stop_after_delay(4),
    )
    # The process gets launched when SQLAlchemy's after_commit event is triggered. This tells us when the COMMIT
    # has been issued to the database, but it does not guarantee that the commit has finished execution.
    # This is why we add some retry logic to ensure that the process is available in the database before continuing.
    def get_process(cls, process_id: int):
        return (
            AsyncProcess.query.filter_by(id=process_id)
            .with_for_update(of=AsyncProcess)
            .populate_existing()
            .one()
        )

    @classmethod
    @with_transaction
    def call_function(cls, process_id):
        """
        Calls the defining function of a given process
        """
        # cls.log(f"Calling function for process_id: {process_id}")
        print("\n")
        logger.info(f"Calling function for process_id {process_id}...")

        process = None

        try:
            from psynet.experiment import get_experiment

            experiment = get_experiment()

            process = cls.get_process(process_id)
            function = process.function

            arguments = cls.preprocess_args(process.arguments)

            timer = time.monotonic()
            process.time_started = datetime.datetime.now()
            db.session.commit()

            call_function(function, **arguments)

            process.time_finished = datetime.datetime.now()
            process.time_taken = time.monotonic() - timer
            process.pending = False
            process.finished = True

            from psynet.trial.main import Trial

            if "self" in arguments and isinstance(arguments["self"], Trial):
                arguments["self"].check_if_can_mark_as_finalized()

        except Exception as err:
            if not isinstance(err, experiment.HandledError):
                experiment.handle_error(err, process=process)

            if process:
                process.pending = False
                process.fail(f"Exception in asynchronous process: {repr(err)}")

        finally:
            db.session.commit()

    @classmethod
    def preprocess_args(cls, arguments):
        """
        Preprocesses the arguments that are passed to the process's function.
        """
        return {key: cls.preprocess_arg(value) for key, value in arguments.items()}

    @classmethod
    def preprocess_arg(cls, arg):
        if isinstance(
            arg, dallinger.db.Base
        ):  # Tests if the object is an SQLAlchemy object
            arg = db.session.merge(arg)  # Reattaches the object to the database session
            db.session.refresh(arg)
        return arg

    @classmethod
    def log(cls, msg):
        raise NotImplementedError

    @classmethod
    def log_to_stdout(cls, msg):
        print(msg)

    @classmethod
    def log_to_redis(cls, msg):
        cls.redis_queue.enqueue_call(
            func=logger.info, args=(), kwargs=dict(msg=msg), timeout=1e10, at_front=True
        )


@event.listens_for(db.session, "after_commit")
def receive_after_commit(session):
    AsyncProcess.launch_all()


class LocalAsyncProcess(AsyncProcess):
    @classmethod
    def launch(cls, process: dict):
        thr = threading.Thread(
            target=cls.thread_function, kwargs={"process_id": process["id"]}
        )
        thr.start()

    @classmethod
    def thread_function(cls, process_id):
        try:
            cls.call_function_with_logger(process_id)
        finally:
            db.session.commit()
            db.session.close()

    @classmethod
    def call_function_with_logger(cls, process_id):
        cls.call_function(process_id)

        # log = io.StringIO()
        # try:
        #     with contextlib.redirect_stdout(log):
        #         # yield
        #         cls.call_function(process_id)
        # finally:
        #     cls.log_to_redis(log.getvalue())

        # with cls.log_output():
        #     cls.call_function(process_id)

    # @classmethod
    # @contextlib.contextmanager
    # def log_output(cls):
    #     log = io.StringIO()
    #     try:
    #         with contextlib.redirect_stdout(log):
    #             yield
    #     finally:
    #         cls.log_to_redis(log.getvalue())

    # @classmethod
    # def log(cls, msg):
    #     cls.log_to_redis(msg)


class WorkerAsyncProcess(AsyncProcess):
    redis_job_id = Column(String)
    timeout = Column(Float)  # note -- currently only applies to non-local proceses
    timeout_scheduled_for = Column(DateTime)
    cancelled = Column(Boolean, default=False)

    def get_launch_spec(self) -> dict:
        spec = super().get_launch_spec()
        spec["timeout"] = self.timeout
        return spec

    def __init__(
        self,
        function,
        arguments=None,
        trial=None,
        participant=None,
        node=None,
        network=None,
        asset=None,
        label=None,
        unique=False,
        timeout=None,  # <-- new argument for this class
    ):
        self.timeout = timeout
        if timeout:
            self.timeout_scheduled_for = datetime.datetime.now() + datetime.timedelta(
                seconds=timeout
            )

        super().__init__(
            function,
            arguments,
            trial=trial,
            participant=participant,
            node=node,
            network=network,
            asset=asset,
            label=label,
            unique=unique,
        )

    @classmethod
    def launch(cls, process: dict):
        # Previously we took the id of the enqueue_call and saved that in Process.redis_job_id,
        # but this is not possible now that the Process object is not accessible.
        cls.redis_queue.enqueue_call(
            func=cls.call_function_with_logger,
            args=(),
            kwargs=dict(process_id=process["id"]),
            timeout=process["timeout"],
        )

    @classmethod
    def check_timeouts(cls):
        processes = cls.query.filter(
            cls.pending,
            ~cls.failed,
            cls.timeout != None,  # noqa -- this is special SQLAlchemy syntax
            cls.timeout_scheduled_for < datetime.datetime.now(),
        ).all()
        for p in processes:
            p.fail(
                "Asynchronous process timed out",
            )
            db.session.commit()

    @property
    def redis_job(self):
        return Job.fetch(self.redis_job_id, connection=redis_conn)

    def cancel(self):
        self.cancelled = True
        self.pending = False
        self.fail("Cancelled asynchronous process")
        self.job.cancel()
        db.session.commit()

    # @classmethod
    # def log(cls, msg):
    #     cls.log_to_stdout(msg)



# pylint: disable=unused-argument,abstract-method

import json
import os
import random
import tempfile
import zipfile
from uuid import uuid4

from markupsafe import Markup, escape

from ..asset import ExperimentAsset
from ..field import claim_var
from ..media import make_batch_file
from ..modular_page import (
    EXTENSIONS,
    AudioSliderControl,
    HtmlSliderControl,
    ImageSliderControl,
    MediaSliderControl,
    ModularPage,
    VideoSliderControl,
)
from ..timeline import MediaSpec
from ..utils import get_logger, linspace
from .gibbs import GibbsNetwork, GibbsNode, GibbsTrial, GibbsTrialMaker

logger = get_logger()


class MediaGibbsNetwork(GibbsNetwork):
    """
    A Network class for Media Gibbs Sampler chains.
    The user should customise this by overriding the attributes
    :attr:`~psynet.trial.media_gibbs.MediaGibbsNetwork.synth_function_location`,
    :attr:`~psynet.trial.media_gibbs.MediaGibbsNetwork.vector_length`,
    :attr:`~psynet.trial.media_gibbs.MediaGibbsNetwork.vector_ranges`,
    and optionally
    :attr:`~psynet.trial.media_gibbs.MediaGibbsNetwork.granularity`,
    :attr:`~psynet.trial.media_gibbs.MediaGibbsNetwork.n_jobs`.
    The user is also invited to override the
    :meth:`psynet.trial.chain.ChainNetwork.make_definition` method
    in situations where different chains are to have different properties
    (e.g. different prompts).

    Attributes
    ----------

    synth_function_location: dict
        A dictionary specifying the function to use for synthesising
        stimuli. The dictionary should contain two arguments:
        one named ``"module_name"``, which identifies by name the module
        in which the function is contained,
        and one named ``"function_name"``, corresponding to the name
        of the function within that module.
        The synthesis function should take three arguments:

            - ``vector``, the parameter vector for the stimulus to be generated.

            - ``output_path``, the output path for the media file to be generated.

            - ``chain_definition``, the ``context`` dictionary for the current chain.

    s3_bucket : str
        Name of the S3 bucket in which the stimuli should be stored.
        The same bucket can be reused between experiments,
        the UUID system used to generate file names should keep them unique.

    vector_length : int
        Must be overridden with the length of the free parameter vector
        that is manipulated during the Gibbs sampling procedure.

    vector_ranges : list
        Must be overridden with a list with length equal to
        :attr:`~psynet.trial.media_gibbs.MediaGibbsNetwork.vector_length`.

    n_jobs : int
        Integer indicating how many parallel processes should be used by an individual worker node
        when generating the stimuli. Note that the final number of parallel processes may
        be considerably more than this; suppose 4 networks are generating stimuli at the same time,
        and we have 3 worker nodes, then the effective number of parallel processes will be 3 x 3 = 9.
        Default is 1, corresponding to no parallelization.

    granularity : Union[int, str]
        When a new :class:`~psynet.trial.media_gibbs.MediaGibbsNode`
        is created, a collection of stimuli are generated that
        span a given dimension of the parameter vector.
        If ``granularity`` is an integer, then this integer sets the number
        of stimuli that are generated, and the stimuli will be spaced evenly
        across the closed interval defined by the corresponding element of
        :attr:`~psynet.trial.media_gibbs.MediaGibbsNetwork.vector_ranges`.
        If ``granularity`` is equal to ``"custom"``, then the spacing of the
        stimuli is instead determined by the media generation function.
    """


class MediaGibbsTrial(GibbsTrial):
    """
    A Trial class for Media Gibbs Sampler chains.
    The user should customise this by overriding the
    :meth:`~psynet.trial.media_gibbs.MediaGibbsTrial.get_prompt`
    method.
    The user must also specify a time estimate
    by overriding the ``time_estimate`` class attribute.
    The user is also invited to override the
    :attr:`~psynet.trial.media_gibbs.MediaGibbsTrial.snap_slider`,
    :attr:`~psynet.trial.media_gibbs.MediaGibbsTrial.autoplay`,
    and
    :attr:`~psynet.trial.media_gibbs.MediaGibbsTrial.minimal_interactions`
    attributes.

    Attributes
    ----------

    snap_slider : bool
        If ``True``, the slider snaps to the location corresponding to the
        closest available media stimulus.
        If ``False`` (default), continuous values are permitted.

    snap_slider_before_release: bool
        If ``True``, the slider snaps to the closest stimulus before release
        rather than after release. This option is only available
        if the stimuli are equally spaced.

    autoplay : bool
        If ``True``, a media corresponding to the initial location on the
        slider will play as soon as the slider is ready for interactions.
        If ``False`` (default), the sound only plays once the participant
        first moves the slider.

    disable_while_playing : bool
        If `True`, the slider is disabled while the media is playing. Default: `False`.

        .. deprecated:: 11.0.0

            Use ``disable_slider_on_change`` instead.

    disable_slider_on_change:
        - ``<float>``: Duration for which the media slider should be disabled after its value changed, in seconds.

        - ``"while_playing"``: The slider will be disabled after a value change, as long as the related media is playing.

        - ``"never"``: The slider will not be disabled after a value change.

        Default: `never`.

    minimal_interactions : int : default: 3
        Minimal interactions with the slider before the user can go to next trial.

    minimal_time : float : default: 3.0
        Minimal amount of time that the user must spend on the page before
        they can proceed to the next trial.

    continuous_updates:
        If `True`, then the slider continuously calls slider-update events when it is dragged,
        rather than just when it is released. In this case the log is disabled. Default: `False`.

    debug : bool
        If ``True``, then the page displays debugging information about the
        current trial. If ``False`` (default), no information is displayed.
        Override this to enable behaviour.

    input_type:
        Defaults to `"HTML5_range_slider"`, which gives a standard horizontal slider.
        The other option currently is `"circular_slider"`, which gives a circular slider.

    random_wrap:
        Defaults to `False`. If `True` then slider is wrapped twice so that there are no boundary jumps.
    """

    time_estimate = None
    snap_slider = False
    snap_slider_before_release = False
    autoplay = False
    disable_while_playing = False
    disable_slider_on_change = "never"
    minimal_interactions = 3
    minimal_time = 3.0
    debug = False
    random_wrap = False
    input_type = "HTML5_range_slider"
    layout = ModularPage.default_layout

    def show_trial(self, experiment, participant):
        self._validate()

        start_value = self.initial_vector[self.active_index]
        vector_range = self.vector_ranges[self.active_index]

        return ModularPage(
            "gibbs_media_trial",
            self._get_prompt(experiment, participant),
            control=MediaSliderControl(
                start_value=start_value,
                min_value=vector_range[0],
                max_value=vector_range[1],
                slider_media=self.media.data[self.network.modality],
                modality=self.network.modality,
                media_locations=self.media_locations,
                autoplay=self.autoplay,
                disable_while_playing=self.disable_while_playing,
                disable_slider_on_change=self.disable_slider_on_change,
                n_steps="n_media" if self.snap_slider_before_release else 10000,
                input_type=self.input_type,
                random_wrap=self.random_wrap,
                reverse_scale=self.reverse_scale,
                directional=False,
                snap_values="media_locations" if self.snap_slider else None,
                minimal_time=self.minimal_time,
                minimal_interactions=self.minimal_interactions,
            ),
            media=self.media,
            time_estimate=self.time_estimate,
        )

    def _get_prompt(self, experiment, participant):
        main = self.get_prompt(experiment, participant)
        if not self.debug:
            return main
        else:
            return (
                (Markup(escape(main)) if isinstance(main, str) else main)
                + Markup("<pre style='overflow: scroll; max-height: 50vh;'>")
                + Markup(escape(json.dumps(self.summarize(), indent=4)))
                + Markup("</pre>")
            )

    def _validate(self):
        if self.snap_slider_before_release and not isinstance(
            self.network.granularity, int
        ):
            raise ValueError(
                "<snap_slider_before_release> can only equal <True> if <granularity> is an integer."
            )
        if (
            self.network.modality in ["image", "html"]
            and self.disable_slider_on_change == "while_playing"
        ):
            raise ValueError(
                f"<disable_slider_on_change> cannot equal <'while_playing'> if the modality is {self.network.modality}."
            )

    @property
    def media(self):
        slider_stimuli = self.slider_stimuli
        media_spec = MediaSpec()

        media_spec.add(
            self.network.modality,
            {
                "slider_stimuli": {
                    "url": slider_stimuli["url"],
                    "ids": [x["id"] for x in slider_stimuli["all"]],
                    "type": "batch",
                    "unzip": self.node.batch_zipped,
                }
            },
        )
        return media_spec

    @property
    def media_locations(self):
        res = {}
        for stimulus in self.slider_stimuli["all"]:
            res[stimulus["id"]] = stimulus["value"]
        return res

    def get_prompt(self, experiment, participant):
        """
        Constructs and returns the prompt to display to the participant.
        This can either be a string of text to display, or raw HTML.
        In the latter case, the HTML should be wrapped in a call to
        ``markupsafe.Markup``.
        """
        raise NotImplementedError

    @property
    def slider_stimuli(self):
        return self.node.slider_stimuli

    @property
    def vector_ranges(self):
        return self.node.vector_ranges


class MediaGibbsNode(GibbsNode):
    """
    A Node class for Media Gibbs sampler chains.
    The user should not have to modify this.
    """

    __extra_vars__ = GibbsNode.__extra_vars__.copy()

    vector_length = 0
    vector_ranges = []
    granularity = 100
    n_jobs = 1
    batch_synthesis = False
    batch_zipped = False

    slider_stimuli = claim_var("slider_stimuli", __extra_vars__)

    def validate(self):
        if not (isinstance(self.vector_length, int) and self.vector_length > 0):
            raise TypeError("<vector_length> must be a positive integer.")

        if not (
            isinstance(self.vector_ranges, list)
            and len(self.vector_ranges) == self.vector_length
        ):
            raise TypeError(
                "<vector_ranges> must be a list with length equal to <vector_length>."
            )

        for r in self.vector_ranges:
            if not (len(r) == 2 and r[0] < r[1]):
                raise ValueError(
                    "Each element of <vector_ranges> must be a list of two numbers in increasing order "
                    "identifying the legal range of the corresponding parameter in the vector."
                )

        if not ((isinstance(self.granularity, int) and self.granularity > 0)):
            raise ValueError("<granularity> must be a positive integer.")

    def random_sample(self, i):
        return random.uniform(self.vector_ranges[i][0], self.vector_ranges[i][1])

    def async_on_deploy(self):
        self.make_stimuli()

    def prepare_stimuli(self, range_to_sample, granularity, output_dir, modality):
        logger.info(modality)
        assert modality in EXTENSIONS.keys()
        values = linspace(range_to_sample[0], range_to_sample[1], granularity)
        ids = [f"slider_stimulus_{_i}" for _i, _ in enumerate(values)]
        files = [f"{_id}" for _id in ids]
        paths = [os.path.join(output_dir, _file) for _file in files]
        stimuli = [
            {"id": _id, "value": _value, "path": _path}
            for _id, _value, _path in zip(ids, values, paths)
        ]
        return values, ids, files, paths, stimuli

    def make_stimuli(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            individual_stimuli_dir = os.path.join(temp_dir, "individual_stimuli")
            os.mkdir(individual_stimuli_dir)

            batch_name = f"{uuid4()}"
            batch_file = f"{batch_name}.batch"
            batch_path = os.path.join(temp_dir, batch_file)
            active_index = self.active_index
            granularity = self.granularity
            vector = self.vector
            range_to_sample = self.vector_ranges[active_index]
            values, ids, files, paths, stimuli = self.prepare_stimuli(
                range_to_sample,
                granularity,
                individual_stimuli_dir,
                self.network.modality,
            )

            if self.batch_synthesis:
                vectors = []

                for idx, value in enumerate(values):
                    _vector = vector.copy()
                    _vector[active_index] = value
                    vectors.append(_vector)

                self.synth_function(
                    vector=vectors,
                    output_path=batch_path,
                    chain_definition=self.context,
                )
            else:

                def _synth(value, path):
                    _vector = vector.copy()
                    _vector[active_index] = value
                    self.synth_function(
                        vector=_vector, output_path=path, chain_definition=self.context
                    )

                parallelize = self.n_jobs > 1
                if parallelize:
                    from joblib import Parallel, delayed

                    logger.info("Using %d processes in parallel" % self.n_jobs)

                    Parallel(n_jobs=self.n_jobs, backend="threading")(
                        delayed(_synth)(_value, _path)
                        for _value, _path in zip(values, paths)
                    )
                else:
                    for _value, _path in zip(values, paths):
                        _synth(_value, _path)

                stimuli = [
                    {"id": _id, "value": _value, "path": _path}
                    for _id, _value, _path in zip(ids, values, paths)
                ]
                self.make_media_batch_file(stimuli, batch_path)

            if self.batch_zipped:
                zipped_batch_file = f"{batch_name}.zip"
                zipped_batch_path = os.path.join(temp_dir, zipped_batch_file)

                with zipfile.ZipFile(zipped_batch_path, mode="w") as archive:
                    archive.write(
                        batch_path,
                        arcname="stim.batch",
                        compress_type=zipfile.ZIP_DEFLATED,
                    )
                batch_path = zipped_batch_path

            asset = ExperimentAsset(
                local_key="slider_stimulus",
                input_path=batch_path,
                parent=self,
            )
            asset.deposit()

            self.slider_stimuli = {"url": asset.url, "all": stimuli}

    @staticmethod
    def make_media_batch_file(stimuli, output_path):
        paths = [x["path"] for x in stimuli]
        make_batch_file(paths, output_path)

    def synth_function(self, vector, output_path, chain_definition=None):
        raise NotImplementedError


class MediaGibbsTrialMaker(GibbsTrialMaker):
    pass


class AudioGibbsNetwork(MediaGibbsNetwork):
    modality = "audio"
    pass


class AudioGibbsTrial(MediaGibbsTrial):
    disable_slider_on_change = "never"

    def show_trial(self, experiment, participant):
        self._validate()

        start_value = self.initial_vector[self.active_index]
        vector_range = self.vector_ranges[self.active_index]
        return ModularPage(
            "gibbs_audio_trial",
            self._get_prompt(experiment, participant),
            control=AudioSliderControl(
                start_value=start_value,
                min_value=vector_range[0],
                max_value=vector_range[1],
                audio=self.media.audio,
                sound_locations=self.media_locations,
                autoplay=self.autoplay,
                disable_while_playing=self.disable_while_playing,
                disable_slider_on_change=self.disable_slider_on_change,
                n_steps="n_media" if self.snap_slider_before_release else 10000,
                input_type=self.input_type,
                random_wrap=self.random_wrap,
                reverse_scale=self.reverse_scale,
                directional=False,
                snap_values="media_locations" if self.snap_slider else None,
                minimal_time=self.minimal_time,
                minimal_interactions=self.minimal_interactions,
            ),
            media=self.media,
            time_estimate=self.time_estimate,
        )


class AudioGibbsNode(MediaGibbsNode):
    pass


class AudioGibbsTrialMaker(MediaGibbsTrialMaker):
    @property
    def default_network_class(self):
        return AudioGibbsNetwork


class ImageGibbsNetwork(MediaGibbsNetwork):
    modality = "image"


class ImageGibbsTrial(MediaGibbsTrial):
    disable_slider_on_change = "never"
    media_width = ""
    media_height = ""
    continuous_updates = False
    layout = ModularPage.default_layout

    def show_trial(self, experiment, participant):
        self._validate()
        if self.continuous_updates and self.disable_slider_on_change != "never":
            raise ValueError(
                "<continuous_updates> can only equal <True> if <disable_slider_on_change> is 'never'."
            )

        start_value = self.initial_vector[self.active_index]
        vector_range = self.vector_ranges[self.active_index]
        return ModularPage(
            f"gibbs_{self.network.modality}_trial",
            self._get_prompt(experiment, participant),
            control=ImageSliderControl(
                start_value=start_value,
                min_value=vector_range[0],
                max_value=vector_range[1],
                slider_media=self.media.data[self.network.modality],
                media_locations=self.media_locations,
                autoplay=self.autoplay,
                disable_slider_on_change=self.disable_slider_on_change,
                media_width=self.media_width,
                media_height=self.media_height,
                n_steps="n_media" if self.snap_slider_before_release else 10000,
                input_type=self.input_type,
                random_wrap=self.random_wrap,
                reverse_scale=self.reverse_scale,
                directional=False,
                snap_values="media_locations" if self.snap_slider else None,
                minimal_time=self.minimal_time,
                minimal_interactions=self.minimal_interactions,
                continuous_updates=self.continuous_updates,
            ),
            media=self.media,
            time_estimate=self.time_estimate,
            layout=self.layout,
        )


class ImageGibbsNode(MediaGibbsNode):
    pass


class ImageGibbsTrialMaker(MediaGibbsTrialMaker):
    @property
    def default_network_class(self):
        return ImageGibbsNetwork


class HtmlGibbsNetwork(MediaGibbsNetwork):
    modality = "html"


class HtmlGibbsTrial(MediaGibbsTrial):
    disable_slider_on_change = "never"
    media_width = ""
    media_height = ""
    continuous_updates = False
    layout = ModularPage.default_layout

    def show_trial(self, experiment, participant):
        self._validate()
        if self.continuous_updates and self.disable_slider_on_change != "never":
            raise ValueError(
                "<continuous_updates> can only equal <True> if <disable_slider_on_change> is 'never'."
            )

        start_value = self.initial_vector[self.active_index]
        vector_range = self.vector_ranges[self.active_index]
        return ModularPage(
            f"gibbs_{self.network.modality}_trial",
            self._get_prompt(experiment, participant),
            control=HtmlSliderControl(
                start_value=start_value,
                min_value=vector_range[0],
                max_value=vector_range[1],
                slider_media=self.media.data[self.network.modality],
                media_locations=self.media_locations,
                autoplay=self.autoplay,
                disable_slider_on_change=self.disable_slider_on_change,
                media_width=self.media_width,
                media_height=self.media_height,
                n_steps="n_media" if self.snap_slider_before_release else 10000,
                input_type=self.input_type,
                random_wrap=self.random_wrap,
                reverse_scale=self.reverse_scale,
                directional=False,
                snap_values="media_locations" if self.snap_slider else None,
                minimal_time=self.minimal_time,
                minimal_interactions=self.minimal_interactions,
                continuous_updates=self.continuous_updates,
            ),
            media=self.media,
            time_estimate=self.time_estimate,
            layout=self.layout,
        )


class HtmlGibbsNode(MediaGibbsNode):
    pass


class HtmlGibbsTrialMaker(MediaGibbsTrialMaker):
    @property
    def default_network_class(self):
        return HtmlGibbsNetwork


class VideoGibbsNetwork(MediaGibbsNetwork):
    modality = "video"


class VideoGibbsTrial(MediaGibbsTrial):
    media_width = ""
    media_height = ""
    disable_slider_on_change = "never"
    layout = ModularPage.default_layout

    def show_trial(self, experiment, participant):
        self._validate()

        start_value = self.initial_vector[self.active_index]
        vector_range = self.vector_ranges[self.active_index]
        return ModularPage(
            f"gibbs_{self.network.modality}_trial",
            self._get_prompt(experiment, participant),
            control=VideoSliderControl(
                start_value=start_value,
                min_value=vector_range[0],
                max_value=vector_range[1],
                slider_media=self.media.data[self.network.modality],
                media_locations=self.media_locations,
                autoplay=self.autoplay,
                disable_while_playing=self.disable_while_playing,
                disable_slider_on_change=self.disable_slider_on_change,
                media_width=self.media_width,
                media_height=self.media_height,
                n_steps="n_media" if self.snap_slider_before_release else 10000,
                input_type=self.input_type,
                random_wrap=self.random_wrap,
                reverse_scale=self.reverse_scale,
                directional=False,
                snap_values="media_locations" if self.snap_slider else None,
                minimal_time=self.minimal_time,
                minimal_interactions=self.minimal_interactions,
            ),
            media=self.media,
            time_estimate=self.time_estimate,
            layout=self.layout,
        )


class VideoGibbsNode(MediaGibbsNode):
    pass


class VideoGibbsTrialMaker(MediaGibbsTrialMaker):
    @property
    def default_network_class(self):
        return VideoGibbsNetwork
